#version 150 core
#pragma optionNV unroll all
//http://stackoverflow.com/questions/12331711/lvalue-in-assignment-too-complex

const int NUMLIGHTS = 6; //!!!!

uniform vec2 gridDimension;

uniform sampler2D albedoAtlas[NUMLIGHTS];
uniform sampler2D normalAtlas[NUMLIGHTS];
uniform sampler2D depthAtlas[NUMLIGHTS];
uniform sampler2D lightAtlas[NUMLIGHTS];

uniform mat4 viewMatrix;
uniform mat4 invViewMatrix;
uniform mat4 invProjMatrix;

uniform mat4 lightView[NUMLIGHTS];
uniform mat4 lightProj;

in vec2 pass_TextCoord;
in vec2 pass_TextCoordAtlas;
//in vec4 shadowCoord[NUMLIGHTS];

in mat4 lightMatrix[NUMLIGHTS];
//in mat4 invLightMatrix[NUMLIGHTS];

out vec4 out_Color;

float LinearizeDepth(sampler2D sampler, vec2 uv) {
	float n = 0.01; // camera z near
	float f = 100.0; // camera z far
	float z = texture(sampler, uv).x;
	return (2.0 * n) / (f + n - z * (f - n));
}

vec4 getWorldSpacePosition(int index, vec2 textCoord) {

	//vec2 tmpCoord = vec2(textCoord.x*gridDimension.x, textCoord.y*gridDimension.y);

	float depth = texture2D(depthAtlas[index], textCoord).x;
	
    vec4 screenSpacePos = vec4(textCoord, depth, 1.0); //tmpCoord
    screenSpacePos -= 0.5;
    screenSpacePos *= 2;
    
    vec4 worldSpace = invViewMatrix * invProjMatrix * screenSpacePos;
    worldSpace / worldSpace.w;
    return worldSpace;
}

vec4 getWorldSpacePositionLight(int lightIndex, vec2 textCoord) {

	//vec2 tmpCoord = vec2(textCoord.x*gridDimension.x, textCoord.y*gridDimension.y);

	float depth = texture2D(depthAtlas[lightIndex+1], textCoord).x;
	
    vec4 screenSpacePos = vec4(textCoord, depth, 1.0); //tmpCoord
    screenSpacePos -= 0.5;
    screenSpacePos *= 2;
    
    mat4 invLight = inverse(lightMatrix[lightIndex]);
    
    vec4 worldSpace = invLight * screenSpacePos;
    worldSpace / worldSpace.w;
    return worldSpace;
}

vec4 calcDirectLight(int lightIndex, vec2 pos, vec3 lightOrigin) {

	int level = 1;
	
	vec4 ws_Position = getWorldSpacePositionLight(lightIndex, pos);
	vec3 normal = textureLod(normalAtlas[1+lightIndex], pos, level).xyz;
	normal *= 2.0;
	normal -= 1.0;

	vec3 lightDir = lightOrigin - ws_Position.xyz;
	normalize(lightDir);

	vec4 color = textureLod(albedoAtlas[1+lightIndex], pos, level);
	
	float intensity = textureLod(lightAtlas[1+lightIndex], pos, level).x;
	
	float ndotl = clamp(dot(lightDir, normal), 0, 1);
	
	return ndotl * color * intensity; //
}

void main(void) {

	out_Color = texture(albedoAtlas[0], pass_TextCoord);

	//out_Color = getWorldSpacePosition(pass_TextCoord);
	//out_Color = texture(depthAtlas[0], pass_TextCoord);

	// WS pixel position
	vec4 ws_Position = getWorldSpacePosition(0, pass_TextCoord);
	//out_Color = ws_Position;

	vec3 normal = texture(normalAtlas[0], pass_TextCoordAtlas).xyz;
	normal *= 2.0;
	normal -= 1.0;
	vec3 lightColor = vec3(0.0);

	float factor = 0.3;//3.0 * 1.0/float(NUMLIGHTS);
	float bias = 0.0025;
	float index = 0.0;
	for (int i=0; i<NUMLIGHTS-1; i++) {
	
		vec4 lightOrigin = inverse(lightMatrix[i])[3];
		lightOrigin /= lightOrigin.w;
		vec3 lightDir = normalize(lightOrigin.xyz - ws_Position.xyz);
	
		vec4 lightPos = lightMatrix[i] * ws_Position; // in light-space
		lightPos /= lightPos.w; // -1 .. 1
		lightPos /= 2.0; // to 0 .. 1
		lightPos += 0.5;
	
		vec2 lookup = lightPos.xy;
		// can be simplified
		if (lightPos.z > texture(depthAtlas[i+1], lookup).x + bias) { // + bias // COMPARE camera space pixel position TO camera depth map
			lightColor += vec3(0.0, 0.0, 0.0);
		} else {
			// outside viewport regions
			if (lookup.x > 1.0 || lookup.y > 1.0 || lookup.x < 0.0 || lookup.y < 0.0) {
				//lightColor += vec3(0.0, 1.0, 1.0);
			} else {
			
				vec4 light = textureLod(lightAtlas[i+1], lightPos.xy, 0); //lightPos
			
				//light.xy = lightPos.xy;
				//light.z = 0;
			
				//lightColor = light.xyz;
				lightColor += factor * clamp(dot(lightDir, normal), 0, 1) * light.xyz;// * texture(lightAtlas[i], lightPos.xy).xyz;
				
				float scale = 10.0;
				for (int y = 0; y < scale; y++) {
					for (int x = 0; x < scale; x++) {
						//lightColor += (1.0/scale * factor * calcDirectLight(i, vec2(x/scale, y/scale))).xyz;
					}
				}
			}
		}
		
		//lightColor = lightDir; //debug
		
	}
	
	out_Color.xyz *= lightColor; // *=
	
	//out_Color = ws_Position;
	
	//out_Color.xyz = texture(lightAtlas[4], pass_TextCoord.xy*4.0).xyz;
	
	if (pass_TextCoordAtlas.x > 0.001) {// avoid too "complex"
		//float depth = texture2D(depthAtlas[0], pass_TextCoord.xy); //LinearizeDepth
		//out_Color = texture2D(depthAtlas[5], pass_TextCoord.xy); //LinearizeDepth
		//out_Color.xyz = vec3(depth, depth, depth);
		
		//out_Color.xyz = texture(albedoAtlas, pass_TextCoord.xy).xyz;
		
		//out_Color = ws_Position;
		
		//out_Color.xy = pass_TextCoord.xy;
		//out_Color.z = 0;
		
		int lightInd = 4;
		
		vec4 lightPos = lightMatrix[lightInd] * ws_Position; // in light-space
		lightPos /= lightPos.w; // -1 .. 1
		lightPos /= 2.0; // to 0 .. 1
		lightPos += 0.5;
		
		out_Color = calcDirectLight(lightInd, pass_TextCoord.xy, lightPos.xyz);
	}
	
}