#version 150 core

const int NUMLIGHTS = 5;

uniform vec2 gridDimension;

uniform sampler2D albedoAtlas;
uniform sampler2D normalAtlas;
uniform sampler2D depthAtlas[NUMLIGHTS];
uniform sampler2D lightAtlas[NUMLIGHTS];

uniform mat4 invViewMatrix;
uniform mat4 invProjMatrix;

in vec2 pass_TextCoord;
in vec2 pass_TextCoordAtlas;
//in vec4 shadowCoord[NUMLIGHTS];

in mat4 lightMatrix[NUMLIGHTS];

out vec4 out_Color;

vec4 getWorldSpacePosition(vec2 textCoord) {

	vec2 tmpCoord = vec2(textCoord.x*gridDimension.x, textCoord.y*gridDimension.y);

	float depth = texture2D(depthAtlas[0], textCoord).x;
    vec4 screenSpacePos = vec4(tmpCoord, depth, 1.0);
    screenSpacePos -= 0.5;
    screenSpacePos *= 2;
    vec4 worldSpace = invViewMatrix * invProjMatrix * screenSpacePos;
    return worldSpace / worldSpace.w;
}

vec4 calcLight(vec2 pos) {
	return texture(lightAtlas[0], pos);
}

void main(void) {

	out_Color = texture(albedoAtlas, pass_TextCoordAtlas); // in single texture

	// WS pixel position
	vec4 ws_Position = getWorldSpacePosition(pass_TextCoordAtlas);
	//out_Color = ws_Position;

	vec3 normal = texture(normalAtlas, pass_TextCoordAtlas).xyz; // in single texture
	normal *= 2.0;
	normal -= 1.0;
	vec3 lightColor = vec3(0.0);

	float factor = 0.3;//3.0 * 1.0/float(NUMLIGHTS);
	float bias = 0.001;
	float index = 0.0;
	for (int i=0; i<NUMLIGHTS; i++) {
	
		vec4 lightOrigin = inverse(lightMatrix[i])[3];
		lightOrigin /= lightOrigin.w;
		vec3 lightDir = normalize(lightOrigin.xyz - ws_Position.xyz);
	
		vec4 lightPos = lightMatrix[i] * ws_Position; // in light-space
		lightPos /= lightPos.w; // -1 .. 1
		
		lightPos /= 2.0; // to 0 .. 1
		lightPos += 0.5;
		
		vec2 lightAtlasPos = lightPos.xy;
		lightAtlasPos.x /= gridDimension.x; // to corresponding part of the atlas
		lightAtlasPos.y /= gridDimension.y;
		lightAtlasPos.x += 1.0/gridDimension.x * mod(i+1,gridDimension.x);
		lightAtlasPos.y += 1.0/gridDimension.y * floor((i+1)/gridDimension.x);
	
		// can be simplified
		if (lightPos.z > texture(depthAtlas[0], lightAtlasPos.xy).x + bias) { // + bias // COMPARE camera space pixel position TO camera depth map
			lightColor += vec3(0.0, 0.0, 0.0);
		} else {
			// outside viewport regions
			if (lightAtlasPos.x > 1.0 || lightAtlasPos.y > 1.0 || lightAtlasPos.x < 0.0 || lightAtlasPos.y < 0.0) {
			
			} else {
			
				vec4 light = textureLod(lightAtlas[i], lightPos.xy, 0);
			
				lightColor += factor * clamp(dot(lightDir, normal), 0, 1) * light.xyz;// * texture(lightAtlas[i], lightPos.xy).xyz;
				
				//float scale = 10.0;
				
				//for (int y = 0; y < scale; y++) {
					//for (int x = 0; x < scale; x++) {
						//lightColor += (1.0/scale * factor * calcLight(vec2(x/scale, y/scale))).xyz;
					//}
				//}
			}
		}
		
		//lightColor = lightDir; //debug
		
	}
	
	out_Color.xyz *= lightColor; // *=
	
	//out_Color.xyz = texture(lightAtlas[4], pass_TextCoord.xy*4.0).xyz;
	//out_Color.xyz = texture(albedoAtlas, pass_TextCoord.xy).xyz;
}