#version 150 core

const int NUMLIGHTS = 5;

uniform vec2 gridDimension;

uniform sampler2D albedoAtlas;
uniform sampler2D normalAtlas;
uniform sampler2D depthAtlas;
uniform sampler2D lightAtlas;

uniform mat4 invViewMatrix;
uniform mat4 invProjMatrix;

in vec2 pass_TextCoord;
//in vec4 shadowCoord[NUMLIGHTS];

in mat4 lightMatrix[NUMLIGHTS];

out vec4 out_Color;

vec4 getWorldSpacePosition(vec2 textCoord) {

	vec2 tmpCoord = vec2(textCoord.x*gridDimension.x, textCoord.y*gridDimension.y);

	float depth = texture2D(depthAtlas, textCoord).x;
    vec4 screenSpacePos = vec4(tmpCoord, depth, 1.0);
    screenSpacePos -= 0.5;
    screenSpacePos *= 2;
    vec4 worldSpace = invViewMatrix * invProjMatrix * screenSpacePos;
    return worldSpace / worldSpace.w;
}

void main(void) {

	out_Color = texture(albedoAtlas, pass_TextCoord);

	//out_Color = getWorldSpacePosition(pass_TextCoord);
	//out_Color = texture(depthAtlas, pass_TextCoord);

	// WS pixel position
	vec4 ws_Position = getWorldSpacePosition(pass_TextCoord);
	//out_Color = ws_Position;

	vec3 normal = texture(normalAtlas, pass_TextCoord).xyz;
	normal *= 2.0;
	normal -= 1.0;
	vec3 lightColor = vec3(0.0);

	float factor = 0.4;//3.0 * 1.0/float(NUMLIGHTS);
	float bias = 0.0001;
	float index = 0.0;
	for (int i=0; i<NUMLIGHTS; i++) {
	
		vec4 lightOrigin = inverse(lightMatrix[i])[3];
		lightOrigin /= lightOrigin.w;
		vec3 lightDir = normalize(lightOrigin.xyz - ws_Position.xyz);
	
		vec4 lightPos = lightMatrix[i] * ws_Position; // in light-space
		lightPos /= lightPos.w; // -1 .. 1
		
		lightPos /= 2.0; // to 0 .. 1
		lightPos += 0.5;
		
		lightPos.x /= gridDimension.x; // to corresponding part of the atlas
		lightPos.y /= gridDimension.y;
		lightPos.x += 1.0/gridDimension.x * mod(i+1,gridDimension.x);
		lightPos.y += 1.0/gridDimension.y * floor((i+1)/gridDimension.x);
	
		// can be simplified
		if (lightPos.z > texture(depthAtlas, lightPos.xy).x + bias) { // + bias // COMPARE camera space pixel position TO camera depth map
			lightColor += vec3(0.0);
		} else {
			// outside viewport regions
			if (lightPos.x > 1.0 || lightPos.y > 1.0 || lightPos.x < 0.0 || lightPos.y < 0.0) {
			
			} else {
				lightColor += factor * clamp(dot(lightDir, normal), 0, 1) * texture(lightAtlas, lightPos.xy).xyz;
			}
		}
		
		//lightColor = lightDir; //debug
		
	}
	
	out_Color.xyz *= lightColor; // *=
	
	//out_Color.xyz = normal;
}