package modules;

import java.util.ArrayList;

/**
 * 
 * Generic double buffer.
 * 
 * Usage:
 *  - On the one side: Object data = new Object(); doubleBuffer.put(data);
 *  - On the other side: doubleBuffer.get(); doubleBuffer.complete()// end of loop
 * 
 * Comment: Java.nio has a double buffer which is the data-type double.
 * 
 * @author DarkMatt
 *
 * @param <T>
 */

public class DoubleBuffer<T> {

	private static final int BUFFERSIZE = 2;
	private ArrayList<T> data;
	
	private int readBuffer;
	private int writeBuffer;
	
	private boolean hasChanged = false;
	
	public DoubleBuffer(T initialData) {
		this.data = new ArrayList<T>(BUFFERSIZE);
		this.data.add(0, initialData); // first element
		this.data.add(1, initialData); // second element
		
		readBuffer = 0;
		writeBuffer = 1;
	}
	
	/**
	 * Writes are atomic -> only a pointer is swapped.
	 * 
	 * @param data
	 */
	public synchronized void put(T data) {
		this.data.set(writeBuffer, data);
		hasChanged = true;
	}
	
	/**
	 * May take some time.
	 * 
	 * @return
	 */
	public T get() {
		return data.get(readBuffer);
	}
	
	public synchronized void complete() {
		if (hasChanged) {
			int tmp = readBuffer;
			readBuffer = writeBuffer;
			writeBuffer = tmp;
			hasChanged = false;
		}
	}
	
}
