package modules;

import modules.commands.CommandExecutor;
import modules.commands.Command_Factory;
import modules.loader.Scene;

public abstract class Module implements Runnable {

	protected Scene scene;
	protected CommandExecutor commandExecutor;
	protected Command_Factory commandFactory;
	protected boolean running;
	protected long frameCounter;
	
	private int updateMs;
	
	/**
	 * A module is a thread.
	 * 
	 * DataObject data ... data to share with other modules
	 * CommandExecutor commandExecutor ... for commands to modules (e.g. shutdown)
	 * Command_Factory commandFactory ... for commands to modules (e.g. shutdown)
	 * 
	 * @param rtm reference to the runtime-manager
	 * @param updateFrequencyMs update rate
	 */
	public Module(Scene scene, CommandExecutor commandExecutor, Command_Factory commandFactory, int updateFrequencyMs) {
		this.scene = scene;
		this.commandExecutor = commandExecutor;
		this.commandFactory = commandFactory;
		running = true;
		frameCounter = 0;
		
		updateMs = updateFrequencyMs;
	}
	
	/**
	 * On initialization inside the thread.
	 */
	protected abstract void init();
	/**
	 * Every frame inside the thread.
	 */
	protected abstract void update();
	/**
	 * On shutdown inside the thread.
	 */
	protected abstract void quit();
	
	@Override
	public void run() {
		
		init();
		
		while(running) {
			
			update();
			
			try {
				Thread.sleep(updateMs);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			
		}
		
		quit();
		
	}
	
	/**
	 * Initialize shutdown.
	 */
	public void shutdown() {
		running = false;
	}
	
}
