package modules.loader;

import java.util.ArrayList;

import org.lwjgl.util.vector.Vector3f;

import modules.loader.loader.LevelLoader;
import modules.loader.objects.Camera;
import modules.loader.objects.DirectionalLight;
import modules.loader.objects.GeneralObject;
import modules.loader.objects.ObjectFactory;
import modules.render.cpu.boundingsphere.BoundingSphere;
import modules.render.objects.ICamera;
import modules.render.objects.IDirectionalLight;
import modules.render.objects.IRenderable;

/**
 * 
 * The scene object is synchronized on writing.
 * 
 * @author DarkMatt
 *
 */

public class Scene {

	public static int[] GRIDDIMENSION = new int[]{3, 2}; // 3 , 2
	public static int NUMLIGHTS = GRIDDIMENSION[0]*GRIDDIMENSION[1]-1;
	
	private int activeScene = 0;
	
	private ICamera camera;
	private ArrayList<IDirectionalLight> lights;
	private ArrayList<GeneralObject> objects;
	private ArrayList<IRenderable> renderObjects;
	
	private ArrayList<IRenderable> screenQuad;
	private ArrayList<IRenderable> debugSpheres;
	
	public Scene() {
		
		camera = new Camera(60, 0.01f, 100.0f);
		camera.rotateX(-30.0f);
		camera.translate(0.0f, 0.0f, 10.0f);
		
		if (activeScene == 0) {
			
			lights = new ArrayList<IDirectionalLight>();
			for (int i = 0; i < NUMLIGHTS; i++) {
				DirectionalLight light = new DirectionalLight(20, 15.0f, 75.0f); //15.0f, 75.0f
				
				light.rotateY(10.0f+(float)i*60.0f); //10.0f+(float)i*60.0f
				light.translate(0.0f, 22.0f, 35.0f);
				light.rotateX(-30.0f);
				/*if (i == 0) {
					light.translate(0.0f, 0.0f, 5.0f);
				}*/
				lights.add(light);
			}
			
		} else if (activeScene == 1) {
			
			camera.translate(0.0f, 0.0f, -1.5f);
			
			// 1 light
			
			GRIDDIMENSION = new int[]{2, 1};
			NUMLIGHTS = GRIDDIMENSION[0]*GRIDDIMENSION[1]-1;
			
			lights = new ArrayList<IDirectionalLight>();
			DirectionalLight light = new DirectionalLight(20, 15.0f, 75.0f); //15.0f, 75.0f
			//light.rotateY(10.0f);
			light.translate(0.0f, 24.0f, 0.0f);
			light.rotateX(-90.0f);
			
			lights.add(light);
			
		} else {
			
			// 1 light
			
			GRIDDIMENSION = new int[]{2, 1};
			NUMLIGHTS = GRIDDIMENSION[0]*GRIDDIMENSION[1]-1;
			
			lights = new ArrayList<IDirectionalLight>();
			DirectionalLight light = new DirectionalLight(20, 15.0f, 75.0f); //15.0f, 75.0f
			light.rotateY(90.0f);
			light.translate(0.0f, 22.0f, 35.0f);
			light.rotateX(-30.0f);
			lights.add(light);
			
		}
		
		objects = new ArrayList<GeneralObject>();
		renderObjects = new ArrayList<IRenderable>();
		screenQuad = new ArrayList<IRenderable>();
		debugSpheres = new ArrayList<IRenderable>();
	}
	
	/**
	 * Adding of objects
	 */
	public synchronized void addObject(GeneralObject object) {
		objects.add(object);
		
		if(object instanceof IRenderable) {
			renderObjects.add((IRenderable)object);
		}
	}
	
	public synchronized void loadTestScene(ObjectFactory factory) {
		
		ArrayList<GeneralObject> world = LevelLoader.loadLevel("/content/level/level" + activeScene + ".xml", "world", factory); // "screenQuad"
		
		for (int i = 0; i < world.size(); i++) {
			objects.add(world.get(i));
			
			if(world.get(i) instanceof IRenderable) {
				renderObjects.add((IRenderable)world.get(i));
			}
		}
		
		ArrayList<GeneralObject> toAdd = LevelLoader.loadLevel("/content/level/level" + activeScene + ".xml", "screenQuad", factory); // "screenQuad"
		
		for (int i = 0; i < toAdd.size(); i++) {
			if(toAdd.get(i) instanceof IRenderable) {
				screenQuad.add((IRenderable)toAdd.get(i));
			}
		}
		
		ArrayList<GeneralObject> boundingSpheres = new ArrayList<GeneralObject>();
		
		GeneralObject s1 = factory.createDefaultSphere();
		((IRenderable)s1).translate(0, 1, 0);
		boundingSpheres.add(s1);
		
		GeneralObject s2 = factory.createDefaultSphere();
		((IRenderable)s2).translate(-3, 1, 1);
		((IRenderable)s2).uniformScale(1.5f);
		boundingSpheres.add(s2);
		
		BoundingSphere b1 = new BoundingSphere(new Vector3f(0, 1, 0), 1.0f);
		BoundingSphere b2 = new BoundingSphere(new Vector3f(-3, 1, 1), 1.5f);
		
		BoundingSphere b3 = new BoundingSphere(b1, b2);
		
		GeneralObject s3 = factory.createDefaultSphere();
		Vector3f p = b3.getPos();
		((IRenderable)s3).translate(p.x, p.y, p.z);
		((IRenderable)s3).uniformScale(b3.getRadius());
		boundingSpheres.add(s3);
		
		for (int i = 0; i < boundingSpheres.size(); i++) {
			if(boundingSpheres.get(i) instanceof IRenderable) {
				debugSpheres.add((IRenderable)boundingSpheres.get(i));
			}
		}
		
	}
	
	public ICamera getCamera() {
		return camera;
	}
	
	public ArrayList<IDirectionalLight> getLights() {
		return lights;
	}
	
	public ArrayList<GeneralObject> getObjects() {
		return objects;
	}
	
	public ArrayList<IRenderable> getRenderObjects() {
		return renderObjects;
	}
	
	public ArrayList<IRenderable> getScreenQuadObjects() {
		return screenQuad;
	}
	
	public ArrayList<IRenderable> getDebugSpheres() {
		return debugSpheres;
	}
	
}
