package modules.loader.loader;

import java.io.IOException;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import modules.loader.objects.GeneralObject;
import modules.loader.objects.ObjectFactory;
import modules.render.objects.IRenderable;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import utility.FileIO;

public class LevelLoader {

	/**
	 * Reads a level file.
	 * 
	 * @param path
	 * @return
	 */
	public static ArrayList<GeneralObject> loadLevel(String path, String levelName, ObjectFactory factory) {
		
		ArrayList<GeneralObject> out = new ArrayList<GeneralObject>();
		
		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder = null;
		try {
			dBuilder = dbFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		}
		Document doc = null;
		try {
			doc = dBuilder.parse(FileIO.getFile(path));
		} catch (SAXException | IOException e) {
			e.printStackTrace();
		}
		doc.getDocumentElement().normalize();
		
		NodeList levels = doc.getElementsByTagName("level");
		
		for (int k = 0; k < levels.getLength(); k++) {
			
			NamedNodeMap levelAttr = levels.item(k).getAttributes();
			
			// find level
			String attribute = levelAttr.item(0).getNodeValue();
			if (attribute.equals(levelName)) {
			
				NodeList objects = levels.item(k).getChildNodes();
				for (int i = 0; i < objects.getLength(); i++) {
					
					if (objects.item(i).hasAttributes() && objects.item(i).getNodeName().equals("object")) {
						NamedNodeMap properties = objects.item(i).getAttributes();
						for (int j = 0; j < properties.getLength(); j++) {
							//System.out.println(properties.item(j).getNodeName() + ": " + properties.item(j).getNodeValue());
							
							if(properties.item(j).getNodeName().equals("model")) {
								out.add(factory.createObject(properties.item(j).getNodeValue()));
							}
							
							if(properties.item(j).getNodeName().equals("position")) {
								IRenderable obj = (IRenderable)(out.get(out.size()-1));
								
								String pos = properties.item(j).getNodeValue();
								String[] positions = pos.split("/");
								
								obj.translate(Float.valueOf(positions[0]), Float.valueOf(positions[1]), Float.valueOf(positions[2]));
							}
							
							if(properties.item(j).getNodeName().equals("rotation")) {
								IRenderable obj = (IRenderable)(out.get(out.size()-1));
								
								String rot = properties.item(j).getNodeValue();
								String[] rotation = rot.split("/");
								
								obj.rotateX(Float.valueOf(rotation[0]));
								obj.rotateY(Float.valueOf(rotation[1]));
								obj.rotateZ(Float.valueOf(rotation[2]));
							}
							
							if(properties.item(j).getNodeName().equals("scale")) {
								IRenderable obj = (IRenderable)(out.get(out.size()-1));
								
								String scale = properties.item(j).getNodeValue();
								
								obj.uniformScale(Float.valueOf(scale));
							}
							
							if(properties.item(j).getNodeName().equals("texture")) {
								IRenderable obj = (IRenderable)(out.get(out.size()-1));
								
								String textureName = properties.item(j).getNodeValue();
								
								obj.attachTexture(textureName);
							}
							
						}
					}
					
					
					
				}
				
			}
			
			
		}
		
		
		
		return out;
		
	}
	
}
