package modules.loader.loader;

import java.io.IOException;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import modules.loader.objects.Camera;
import modules.render.objects.ICamera;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import utility.FileIO;

public class ViewportXMLSave {

	public static void saveObjectDataToXML(ArrayList<ICamera> vieports) throws Exception {
		
		DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder = docFactory.newDocumentBuilder();
 
		// root elements
		Document doc = docBuilder.newDocument();
		Element rootElement = doc.createElement("Data");
		doc.appendChild(rootElement);
 
		for (int i = 0; i < vieports.size(); i++) {
			// staff elements
			Element viewport = doc.createElement("Viewport");
			rootElement.appendChild(viewport);
			
			Attr attr = doc.createAttribute("id");
			attr.setValue(String.valueOf(i));
			viewport.setAttributeNode(attr);
	 
			// viewMatrix elements
			Element viewMatrix = doc.createElement("viewMatrix");
			viewMatrix.appendChild(doc.createTextNode(matrixToString(vieports.get(i).getViewMatrix())));
			viewport.appendChild(viewMatrix);
			
			// projMatrix elements
			Element projMatrix = doc.createElement("projMatrix");
			projMatrix.appendChild(doc.createTextNode(matrixToString(vieports.get(i).getProjMatrix())));
			viewport.appendChild(projMatrix);
			
			// inverse projMatrix elements
			Element invProjMatrix = doc.createElement("invProjMatrix");
			invProjMatrix.appendChild(doc.createTextNode(matrixToString(vieports.get(i).getInvProjMatrix())));
			viewport.appendChild(invProjMatrix);
			
		}
 
		// write the content into xml file
		TransformerFactory transformerFactory = TransformerFactory.newInstance();
		Transformer transformer = transformerFactory.newTransformer();
		DOMSource source = new DOMSource(doc);
		StreamResult result = new StreamResult(FileIO.getFile("/content/output/data.xml"));
		
 
		transformer.transform(source, result);
		
	}
	
	public static ArrayList<ICamera> loadObjectDataFromXML() throws Exception {
		
		ArrayList<ICamera> out = new ArrayList<ICamera>();
		
		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder = null;
		try {
			dBuilder = dbFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		}
		Document doc = null;
		try {
			doc = dBuilder.parse(FileIO.getFile("/content/output/data.xml"));
		} catch (SAXException | IOException e) {
			e.printStackTrace();
		}
		doc.getDocumentElement().normalize();
		
		NodeList viewports = doc.getElementsByTagName("Viewport");
		
		//System.out.println("Number of viewports " + viewports.getLength());
		
		for (int i=0; i < viewports.getLength(); i++) {
			
			Node object = viewports.item(i);
			
			if (object.getNodeType() == Node.ELEMENT_NODE) {
				
				Camera toAdd = new Camera(0, 0, 0);
				
				Element objectElement = (Element)object;
				
				NodeList vmList = objectElement.getElementsByTagName("viewMatrix");
                Element vmElement = (Element)vmList.item(0);
                NodeList vmvalueList = vmElement.getChildNodes();
                String matrix = ((Node)vmvalueList.item(0)).getNodeValue().trim();
                String[] components = matrix.split("[/]");
                float[] viewMatrix = new float[components.length];
                for (int j = 0; j < viewMatrix.length; j++) {
                	viewMatrix[j] = Float.valueOf(components[j]);
				}
                
                NodeList pmList = objectElement.getElementsByTagName("projMatrix");
                Element pmElement = (Element)pmList.item(0);
                NodeList pmvalueList = pmElement.getChildNodes();
                matrix = ((Node)pmvalueList.item(0)).getNodeValue().trim();
                components = matrix.split("[/]");
                float[] projMatrix = new float[components.length];
                for (int j = 0; j < projMatrix.length; j++) {
                	projMatrix[j] = Float.valueOf(components[j]);
				}
                
                NodeList ipmList = objectElement.getElementsByTagName("invProjMatrix");
                Element ipmElement = (Element)ipmList.item(0);
                NodeList ipmvalueList = ipmElement.getChildNodes();
                matrix = ((Node)ipmvalueList.item(0)).getNodeValue().trim();
                components = matrix.split("[/]");
                float[] invProjMatrix = new float[components.length];
                for (int j = 0; j < invProjMatrix.length; j++) {
                	invProjMatrix[j] = Float.valueOf(components[j]);
				}
                
                toAdd.setViewMatrix(viewMatrix);
                toAdd.setViewMatrix(projMatrix);
                toAdd.setViewMatrix(invProjMatrix);
                
                out.add(toAdd);
                
			}
			
		}
		
		return out;
		
	}
	
	private static String matrixToString(float[] matrix) {
		String out = "";
		
		for (int i = 0; i < matrix.length-1; i++) {
			out += matrix[i] + "/";
		}
		out += matrix[matrix.length-1];
		
		return out;
	}
	
}
