package modules.loader.objects;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;

import modules.render.objects.ICamera;

public class Camera implements ICamera {

	/** Camera matrix - inverse view matrix */
	private Matrix4f cameraMatrix;
	
	private Matrix4f projMatrix;
	private Matrix4f invProjMatrix;
	
	public Camera(float fovy, float zNear, float zFar) {
		cameraMatrix = new Matrix4f();
		projMatrix = new Matrix4f();
		invProjMatrix = new Matrix4f();
		
		this.set_projection(fovy, 1, zNear, zFar);
	}
	
	//based on gluPersective
	private void set_projection(float fovy, float aspect, float zNear, float zFar) {
			
		float f = 1.0f / (float)Math.tan(Math.toRadians(fovy)/2.0f);
		
		projMatrix.setIdentity();
		
		projMatrix.m00 = f/aspect;
		projMatrix.m11 = f;
		projMatrix.m22 = (zFar+zNear)/(zNear-zFar);
		projMatrix.m32 = (2*zFar*zNear)/(zNear-zFar);
		projMatrix.m23 = -1.0f;
		projMatrix.m33 = 0.0f;
		
		invProjMatrix.m00 = aspect/f;
		invProjMatrix.m11 = 1.0f/f;
		invProjMatrix.m22 = 0.0f;
		invProjMatrix.m32 = -1.0f;
		invProjMatrix.m23 = (zNear-zFar)/(2*zFar*zNear);
		invProjMatrix.m33 = (zNear+zFar)/(2*zFar*zNear);
		
	}
	
	@Override
	public float[] getCameraMatrix() {
		float[] out = 
			{
				cameraMatrix.m00, cameraMatrix.m10, cameraMatrix.m20, cameraMatrix.m30,
				cameraMatrix.m01, cameraMatrix.m11, cameraMatrix.m21, cameraMatrix.m31,
				cameraMatrix.m02, cameraMatrix.m12, cameraMatrix.m22, cameraMatrix.m32,
				cameraMatrix.m03, cameraMatrix.m13, cameraMatrix.m23, cameraMatrix.m33
			};
		return out;
	}

	@Override
	public float[] getViewMatrix() {
		
		Matrix4f invView = new Matrix4f();
		
		Matrix4f.invert(cameraMatrix, invView);
		
		float[] out = 
			{
				invView.m00, invView.m10, invView.m20, invView.m30,
				invView.m01, invView.m11, invView.m21, invView.m31,
				invView.m02, invView.m12, invView.m22, invView.m32,
				invView.m03, invView.m13, invView.m23, invView.m33
			};
		return out;
	}
	
	@Override
	public float[] getProjMatrix() {
		float[] out = 
			{
				projMatrix.m00, projMatrix.m10, projMatrix.m20, projMatrix.m30,
				projMatrix.m01, projMatrix.m11, projMatrix.m21, projMatrix.m31,
				projMatrix.m02, projMatrix.m12, projMatrix.m22, projMatrix.m32,
				projMatrix.m03, projMatrix.m13, projMatrix.m23, projMatrix.m33
			};
		return out;
	}
	
	@Override
	public float[] getInvProjMatrix() {
		float[] out = 
			{
				invProjMatrix.m00, invProjMatrix.m10, invProjMatrix.m20, invProjMatrix.m30,
				invProjMatrix.m01, invProjMatrix.m11, invProjMatrix.m21, invProjMatrix.m31,
				invProjMatrix.m02, invProjMatrix.m12, invProjMatrix.m22, invProjMatrix.m32,
				invProjMatrix.m03, invProjMatrix.m13, invProjMatrix.m23, invProjMatrix.m33
			};
		return out;
	}

	@Override
	public void translate(float x, float y, float z) {
		cameraMatrix.translate(new Vector3f(x,y,z));
	}

	@Override
	public void rotateX(float angle) {
		float toDegree = angle * (float)Math.PI/180.0f;
		cameraMatrix.rotate(toDegree, new Vector3f(1.0f, 0.0f, 0.0f));
	}

	@Override
	public void rotateY(float angle) {
		float toDegree = angle * (float)Math.PI/180.0f;
		cameraMatrix.rotate(toDegree, new Vector3f(0.0f, 1.0f, 0.0f));
	}

	@Override
	public void rotateZ(float angle) {
		float toDegree = angle * (float)Math.PI/180.0f;
		cameraMatrix.rotate(toDegree, new Vector3f(0.0f, 0.0f, 1.0f));
	}

	@Override
	public void setViewMatrix(float[] matrix) {
		cameraMatrix.m00 = matrix[0];  cameraMatrix.m10 = matrix[1];  cameraMatrix.m20 = matrix[2];  cameraMatrix.m30 = matrix[3];
		cameraMatrix.m01 = matrix[4];  cameraMatrix.m11 = matrix[5];  cameraMatrix.m21 = matrix[6];  cameraMatrix.m31 = matrix[7];
		cameraMatrix.m02 = matrix[8];  cameraMatrix.m12 = matrix[9];  cameraMatrix.m22 = matrix[10]; cameraMatrix.m32 = matrix[11];
		cameraMatrix.m03 = matrix[12]; cameraMatrix.m13 = matrix[13]; cameraMatrix.m23 = matrix[14]; cameraMatrix.m33 = matrix[15];
	}

	@Override
	public void setProjMatrix(float[] matrix) {
		projMatrix.m00 = matrix[0];  projMatrix.m10 = matrix[1];  projMatrix.m20 = matrix[2];  projMatrix.m30 = matrix[3];
		projMatrix.m01 = matrix[4];  projMatrix.m11 = matrix[5];  projMatrix.m21 = matrix[6];  projMatrix.m31 = matrix[7];
		projMatrix.m02 = matrix[8];  projMatrix.m12 = matrix[9];  projMatrix.m22 = matrix[10]; projMatrix.m32 = matrix[11];
		projMatrix.m03 = matrix[12]; projMatrix.m13 = matrix[13]; projMatrix.m23 = matrix[14]; projMatrix.m33 = matrix[15];
	}

	@Override
	public void setInvProjMatrix(float[] matrix) {
		invProjMatrix.m00 = matrix[0];  invProjMatrix.m10 = matrix[1];  invProjMatrix.m20 = matrix[2];  invProjMatrix.m30 = matrix[3];
		invProjMatrix.m01 = matrix[4];  invProjMatrix.m11 = matrix[5];  invProjMatrix.m21 = matrix[6];  invProjMatrix.m31 = matrix[7];
		invProjMatrix.m02 = matrix[8];  invProjMatrix.m12 = matrix[9];  invProjMatrix.m22 = matrix[10]; invProjMatrix.m32 = matrix[11];
		invProjMatrix.m03 = matrix[12]; invProjMatrix.m13 = matrix[13]; invProjMatrix.m23 = matrix[14]; invProjMatrix.m33 = matrix[15];
	}

}
