package modules.loader.objects;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.lwjgl.opengl.GL11;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;

import utility.BufferConverter;

import modules.render.objects.VBOArrayObject;
import modules.render.objects.VBOArrayObject.Location;

public class ObjectFactory {

	public GeneralObject createObject(String model) {
		
		if (model.equals("defaultTriangle")) {
			return createDefaultTriangle();
		} else if (model.equals("defaultQuad")) {
			return createDefaultQuad();
		} else if (model.equals("defaultQuadTextured")) {
			return createDefaultQuadTextured();
		} else if (model.equals("defaultCube")) {
			return createDefaultCube();
		} else if (model.equals("defaultSphere")) {
			return createDefaultSphere();
		} else {	
			return createModel(model);
		}
	}
	
	public GeneralObject createDefaultTriangle() {
		
		SceneObject obj = new SceneObject();
		
		float[] vPos = new float[]
			{
				-1.0f,  1.0f, 0f,	//0
				-1.0f, -1.0f, 0f,	//1
				1.0f, -1.0f, 0f	//2
			};
		float[] vColor = new float[]
			{
				1.0f, 0.0f, 0.0f,
				0.0f, 1.0f, 0.0f,
				0.0f, 1.0f, 1.0f
			};
		int[] index = new int[]{0, 1, 2};
		
		ArrayList<VBOArrayObject> vertexData = new ArrayList<VBOArrayObject>();
		VBOArrayObject toAdd = new VBOArrayObject(BufferConverter.toBuffer(vPos), Location.DEFAULT_VERTEX_POSITION, GL11.GL_FLOAT, 3); // positions
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(vColor), Location.DEFAULT_COLOR_POSITION, GL11.GL_FLOAT, 3); // colors
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(index), Location.DEFAULT_INDEX_POSITION, GL11.GL_FLOAT, index.length); // indices
		vertexData.add(toAdd);
		
		obj.attachVertexData(vertexData);
		
		return obj;
	}
	
	public GeneralObject createDefaultQuad() {
		
		SceneObject obj = new SceneObject();
		
		float[] vPos = new float[]
			{
				-1.0f,  1.0f, 0f,	//0
				-1.0f, -1.0f, 0f,	//1
				 1.0f, -1.0f, 0f,	//2
				 1.0f,  1.0f, 0f	//3
			};
		float[] vColor = new float[]
			{
				1.0f, 0.0f, 0.0f,
				1.0f, 1.0f, 0.0f,
				1.0f, 0.0f, 1.0f,
				1.0f, 0.0f, 0.0f
			};
		int[] index = new int[]{0, 1, 2, 2, 3, 0};
		
		ArrayList<VBOArrayObject> vertexData = new ArrayList<VBOArrayObject>();
		VBOArrayObject toAdd = new VBOArrayObject(BufferConverter.toBuffer(vPos), Location.DEFAULT_VERTEX_POSITION, GL11.GL_FLOAT, 3); // positions
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(vColor), Location.DEFAULT_COLOR_POSITION, GL11.GL_FLOAT, 3); // colors
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(index), Location.DEFAULT_INDEX_POSITION, GL11.GL_FLOAT, index.length); // indices
		vertexData.add(toAdd);
		
		obj.attachVertexData(vertexData);
		
		return obj;
	}
	
	public GeneralObject createDefaultQuadTextured() {
		
		SceneObject obj = new SceneObject();
		
		float[] vPos = new float[]
			{
				-1.0f,  1.0f, 0f,	//0
				-1.0f, -1.0f, 0f,	//1
				 1.0f, -1.0f, 0f,	//2
				 1.0f,  1.0f, 0f	//3
			};
		float[] vTexCoord = new float[]
			{
				0.0f, 1.0f,
				0.0f, 0.0f,
				1.0f, 0.0f,
				1.0f, 1.0f
			};
		int[] index = new int[]{0, 1, 2, 2, 3, 0};
		
		ArrayList<VBOArrayObject> vertexData = new ArrayList<VBOArrayObject>();
		VBOArrayObject toAdd = new VBOArrayObject(BufferConverter.toBuffer(vPos), Location.DEFAULT_VERTEX_POSITION, GL11.GL_FLOAT, 3); // positions
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(vTexCoord), Location.DEFAULT_TEXTCOORD_POSITION, GL11.GL_FLOAT, 2); // textcords
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(index), Location.DEFAULT_INDEX_POSITION, GL11.GL_FLOAT, index.length); // indices
		vertexData.add(toAdd);
		
		obj.attachVertexData(vertexData);
		
		return obj;
	}
	
	public GeneralObject createDefaultCube() {
		
		SceneObject obj = new SceneObject();
		
		float[] vPos = new float[]
			{
				-1.0f, -1.0f,  1.0f,
				 1.0f, -1.0f,  1.0f,
				 1.0f,  1.0f,  1.0f,
				-1.0f,  1.0f,  1.0f,
				-1.0f, -1.0f, -1.0f,
				 1.0f, -1.0f, -1.0f,
				 1.0f,  1.0f, -1.0f,
				-1.0f,  1.0f, -1.0f
			};
		float[] vColor = new float[]
			{
				0.7f, 0.7f, 0.7f,
				0.7f, 0.7f, 0.7f,
				0.7f, 0.7f, 0.7f,
				0.7f, 0.7f, 0.7f,
				0.7f, 0.7f, 0.7f,
				0.7f, 0.7f, 0.7f,
				0.7f, 0.7f, 0.7f,
				0.7f, 0.7f, 0.7f,
				0.7f, 0.7f, 0.7f
			};
		int[] index = new int[]
			{
				0, 1, 2, 2, 3, 0, 
				3, 2, 6, 6, 7, 3, 
				7, 6, 5, 5, 4, 7, 
				4, 0, 3, 3, 7, 4, 
				5, 1, 0, 0, 4, 5,
				1, 5, 6, 6, 2, 1 
			};
		
		ArrayList<VBOArrayObject> vertexData = new ArrayList<VBOArrayObject>();
		VBOArrayObject toAdd = new VBOArrayObject(BufferConverter.toBuffer(vPos), Location.DEFAULT_VERTEX_POSITION, GL11.GL_FLOAT, 3); // positions
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(vColor), Location.DEFAULT_COLOR_POSITION, GL11.GL_FLOAT, 3); // colors
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(index), Location.DEFAULT_INDEX_POSITION, GL11.GL_FLOAT, index.length); // indices
		vertexData.add(toAdd);
		
		obj.attachVertexData(vertexData);
		
		return obj;
	}
	
	public GeneralObject createDefaultSphere() {
		return createModel("/models/sphere.obj");
	}
	
	public GeneralObject createModel(String path) {
		
		SceneObject obj = new SceneObject();
		
		float[] vertices = null;
		float[] colors = null;
		float[] textCoords = null;
		float[] normals = null;
		float[] tangents = null;
		int[] indices = null;
		
		
		String line = "";
		URI uri = null;
		try {
			uri = new URI("file://"+ObjectFactory.class.getProtectionDomain().getCodeSource().getLocation().getPath());
		} catch (URISyntaxException e1) {
			throw new RuntimeException("URI fault");
		}
		
        File f = new File(uri);
		BufferedReader reader;
		
		ArrayList<Float> v = new ArrayList<Float>();
		ArrayList<Float> vt = new ArrayList<Float>();
		ArrayList<Float> vn = new ArrayList<Float>();
		ArrayList<Integer> iv = new ArrayList<Integer>();
		ArrayList<Integer> ivt = new ArrayList<Integer>();
		ArrayList<Integer> ivn = new ArrayList<Integer>();
		
		try {
			reader = new BufferedReader(new FileReader(f.getParent() + "/content" + path));
			while( (line=reader.readLine()) != null){
				if (line.length() > 2 && line.substring(0, 2).equals("v ")) {
					String[] p = line.split("\\s+");
					v.add(Float.valueOf(p[1]));
					v.add(Float.valueOf(p[2]));
					v.add(Float.valueOf(p[3]));
				}
				if (line.length() > 2 && line.substring(0, 2).equals("vt")) {
					String[] p = line.split("\\s+");
					vt.add(Float.valueOf(p[1]));
					vt.add(Float.valueOf(p[2]));
				}
				if (line.length() > 2 && line.substring(0, 2).equals("vn")) {
					String[] p = line.split("\\s+");
					vn.add(Float.valueOf(p[1]));
					vn.add(Float.valueOf(p[2]));
					vn.add(Float.valueOf(p[3]));
				}
				if (line.length() > 2 && line.substring(0, 2).equals("f ")) {
					//indices: v/vt/vn (should be all equal!)
					String[] p = line.split("\\s+");
					for (int i=1; i<4; i++) {
						String tmp = p[i];
						String[] pts = tmp.split("\\/");
						iv.add(Integer.valueOf(pts[0])-1); //ATTENTION: in *.obj file the indices start at 1
						ivt.add(Integer.valueOf(pts[1])-1);
						ivn.add(Integer.valueOf(pts[2])-1);
					}
				}
			}
			reader.close();
		} catch (Exception e) {
			throw new RuntimeException("Couldnt load model! Was looking for: " + f.getParent()  + "/content" + path +
					" \nReason: " + e);
		}
		
		//advantage.. IO is reduced.. (Slow)
		//modify the arrays to make iv == ivt == ivn
		//PRECONDITION(!): the v/vt/vn order is not arbitrary (modeling programs do not create those)
		
		ArrayList<Float> vertex = new ArrayList<Float>();
		ArrayList<Float> color = new ArrayList<Float>();
		ArrayList<Float> textCoord = new ArrayList<Float>();
		ArrayList<Float> normal = new ArrayList<Float>();
		ArrayList<Integer> index = new ArrayList<Integer>();
		Map<Integer,String> hm = new HashMap<Integer,String>();
		
		for (int i=0; i<iv.size(); i++) {
			String tmp = iv.get(i)+"/"+ivt.get(i)+"/"+ivn.get(i);
			if (!hm.containsValue(tmp)) {
				hm.put(hm.size(), tmp);
				int vPos = iv.get(i);
				vertex.add(v.get(3*vPos+0));
				vertex.add(v.get(3*vPos+1));
				vertex.add(v.get(3*vPos+2));
				color.add(1.0f);
				color.add(1.0f);
				color.add(1.0f);
				textCoord.add(vt.get(2*ivt.get(i)+0));
				textCoord.add(vt.get(2*ivt.get(i)+1));
				normal.add(vn.get(3*ivn.get(i)+0));
				normal.add(vn.get(3*ivn.get(i)+1));
				normal.add(vn.get(3*ivn.get(i)+2));
				index.add(hm.size()-1);
			} else {
				int key = 0;
				for (Entry<Integer, String> entry : hm.entrySet()) {
			        if (tmp.equals(entry.getValue())) {
			        	key = entry.getKey();
			        }
			    }
				index.add(key);
			}
		}
		
		vertices = new float[vertex.size()];
		for (int i=0; i<vertex.size(); i++) vertices[i] = vertex.get(i);
		colors = new float[color.size()];
		for (int i=0; i<color.size(); i++) colors[i] = color.get(i);
		textCoords = new float[textCoord.size()];
		for (int i=0; i<textCoord.size(); i++) textCoords[i] = textCoord.get(i);
		normals = new float[normal.size()];
		for (int i=0; i<normal.size(); i++) normals[i] = normal.get(i);
		indices = new int[index.size()];
		for (int i=0; i<index.size(); i++) indices[i] = index.get(i);
		
		// calculate tangents
		tangents = calculateTangentArray(vertices.length/3, vertices, normals, textCoords, indices.length/3, indices);
		
		/*//get bs radius
		for (int i = 0; i < vertices.length/3; i++) {
			float dist = (float)Math.sqrt(vertices[3*i+0]*vertices[3*i+0] + vertices[3*i+1]*vertices[3*i+1] + vertices[3*i+2]*vertices[3*i+2]);
			if (dist > bsRadius) {
				bsRadius = dist;
			}
		}*/
		
		ArrayList<VBOArrayObject> vertexData = new ArrayList<VBOArrayObject>();
		VBOArrayObject toAdd = new VBOArrayObject(BufferConverter.toBuffer(vertices), Location.DEFAULT_VERTEX_POSITION, GL11.GL_FLOAT, 3); // positions
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(normals), Location.DEFAULT_NORMAL_POSITION, GL11.GL_FLOAT, 3); // normals
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(textCoords), Location.DEFAULT_TEXTCOORD_POSITION, GL11.GL_FLOAT, 2); // texture coords
		vertexData.add(toAdd);
		toAdd = new VBOArrayObject(BufferConverter.toBuffer(indices), Location.DEFAULT_INDEX_POSITION, GL11.GL_FLOAT, indices.length); // indices
		vertexData.add(toAdd);
		
		obj.attachVertexData(vertexData);
		
		return obj;
	}
	
	/**
	 * Taken from http://www.terathon.com/code/tangent.html
	 * 
	 * @param vertexCount
	 * @param vertices
	 * @param normals
	 * @param textCoords
	 * @param triangleCount
	 * @param indices
	 */
	private float[] calculateTangentArray(int vertexCount, float[] vertices, float[] normals, float[] textCoords, int triangleCount, int[] indices) {
		
		float[] tan1 = new float[3*vertexCount];
		float[] tan2 = new float[3*vertexCount];
	    
	    for (int a = 0; a < triangleCount; a++) {
	        int i1 = indices[3*a+0];	// index to first triangle vertex
	        int i2 = indices[3*a+1];	// index to second triangle vertex
	        int i3 = indices[3*a+2];	// index to third triangle vertex
	        
	        Vector3f v1 = new Vector3f(vertices[3*i1+0], vertices[3*i1+1], vertices[3*i1+2]);
	        Vector3f v2 = new Vector3f(vertices[3*i2+0], vertices[3*i2+1], vertices[3*i2+2]);
	        Vector3f v3 = new Vector3f(vertices[3*i3+0], vertices[3*i3+1], vertices[3*i3+2]);
	        
	        Vector2f w1 = new Vector2f(textCoords[2*i1+0], textCoords[2*i1+1]);
	        Vector2f w2 = new Vector2f(textCoords[2*i2+0], textCoords[2*i2+1]);
	        Vector2f w3 = new Vector2f(textCoords[2*i3+0], textCoords[2*i3+1]);
	        
	        float x1 = v2.x - v1.x;
	        float x2 = v3.x - v1.x;
	        float y1 = v2.y - v1.y;
	        float y2 = v3.y - v1.y;
	        float z1 = v2.z - v1.z;
	        float z2 = v3.z - v1.z;
	        
	        float s1 = w2.x - w1.x;
	        float s2 = w3.x - w1.x;
	        float t1 = w2.y - w1.y;
	        float t2 = w3.y - w1.y;
	        
	        float r = 1.0F / (s1 * t2 - s2 * t1);
	        Vector3f sdir = new Vector3f((t2 * x1 - t1 * x2) * r, (t2 * y1 - t1 * y2) * r, (t2 * z1 - t1 * z2) * r);
	        Vector3f tdir = new Vector3f((s1 * x2 - s2 * x1) * r, (s1 * y2 - s2 * y1) * r, (s1 * z2 - s2 * z1) * r);
	        
	        tan1[3*i1+0] = sdir.x;	//first vertex tangent
	        tan1[3*i1+1] = sdir.y;
	        tan1[3*i1+2] = sdir.z;
	        tan1[3*i2+0] = sdir.x;	//second vertex tangent
	        tan1[3*i2+1] = sdir.y;
	        tan1[3*i2+2] = sdir.z;
	        tan1[3*i3+0] = sdir.x;	//third vertex tangent
	        tan1[3*i3+1] = sdir.y;
	        tan1[3*i3+2] = sdir.z;
	        
	        tan2[3*i1+0] = tdir.x;	//first vertex tangent
	        tan2[3*i1+1] = tdir.y;
	        tan2[3*i1+2] = tdir.z;
	        tan2[3*i2+0] = tdir.x;	//second vertex tangent
	        tan2[3*i2+1] = tdir.y;
	        tan2[3*i2+2] = tdir.z;
	        tan2[3*i3+0] = tdir.x;	//third vertex tangent
	        tan2[3*i3+1] = tdir.y;
	        tan2[3*i3+2] = tdir.z;
	    }
	    
	    float[] tangents = new float[4*vertexCount];
	    
	    for (int a = 0; a < vertexCount; a++)
	    {
	        Vector3f n = new Vector3f(normals[3*a+0], normals[3*a+1], normals[3*a+2]);
	        Vector3f t = new Vector3f(tan1[3*a+0], tan1[3*a+1], tan1[3*a+2]);
	        
	        // Gram-Schmidt orthogonalize //(t - n * Dot(n, t)).Normalize()
	        Vector3f result = new Vector3f();
	        Vector3f.sub(t, (Vector3f)n.scale(Vector3f.dot(n, t)), result).normalise();
	        tangents[4*a+0] = result.x;
	        tangents[4*a+1] = result.y;
	        tangents[4*a+2] = result.z;
	        
	        // Calculate handedness
	        Vector3f tan2V = new Vector3f(tan2[3*a], tan2[3*a+1], tan2[3*a+2]);
	        Vector3f.cross(n, t, result);
	        tangents[4*a+3] = (Vector3f.dot(result, tan2V) < 0.0F) ? -1.0F : 1.0F;
	    }
	    
	    return tangents;
	    
	}
	
}
