package modules.loader.objects;

import java.util.ArrayList;

import org.lwjgl.opengl.GL11;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;

import utility.BufferConverter;

import modules.render.Shader;
import modules.render.TextureCollection;
import modules.render.VBO;
import modules.render.objects.IRenderable;
import modules.render.objects.VBOArrayObject;

public class SceneObject extends GeneralObject implements IRenderable {

	private int objectIndex;
	private ArrayList<VBOArrayObject> vertexData;
	private int textureIndex;
	private String textureName;
	
	private Matrix4f modelMatrix;
	
	
	private boolean vertexDataIsUploaded;
	
	public SceneObject() {
		textureName = "";
		textureIndex = -1;
		modelMatrix = new Matrix4f();
	}
	
	@Override
	public void attachVertexData(ArrayList<VBOArrayObject> vertexData) {
		this.vertexData = vertexData;
		vertexDataIsUploaded = false;
	}

	@Override
	public void uploadVertexData(VBO vbo) {
		objectIndex = vbo.init(vertexData);
		vertexDataIsUploaded = true;
	}

	@Override
	public void attachTexture(String textureName) {
		this.textureName = textureName;
	}
	
	@Override
	public void uploadTexture(TextureCollection textures) {
		if (!textureName.equals("")) {
			textureIndex = textures.uploadImage("/content"+textureName);
		}
	}
	
	@Override
	public void translate(float x, float y, float z) {
		modelMatrix.translate(new Vector3f(x, y, z));
	}
	
	@Override
	public void rotateX(float angle) {
		float toDegree = angle * (float)Math.PI/180.0f;
		modelMatrix.rotate(toDegree, new Vector3f(1.0f, 0.0f, 0.0f));
	}

	@Override
	public void rotateY(float angle) {
		float toDegree = angle * (float)Math.PI/180.0f;
		modelMatrix.rotate(toDegree, new Vector3f(0.0f, 1.0f, 0.0f));
	}

	@Override
	public void rotateZ(float angle) {
		float toDegree = angle * (float)Math.PI/180.0f;
		modelMatrix.rotate(toDegree, new Vector3f(0.0f, 0.0f, 1.0f));
	}
	
	@Override
	public void draw(Shader currentShader, VBO vbo, TextureCollection textures) {
		
		if (textureIndex != -1) {
			textures.bindTexture(textureIndex);
		}
		
		// BIND UNIFORMS HERE
		
		currentShader.initUniform("modelMatrix"); // TODO redundant
		currentShader.setValueMat4f("modelMatrix", BufferConverter.toBuffer(modelMatrix));
		
		vbo.draw(objectIndex, GL11.GL_TRIANGLES, vertexData.get(vertexData.size()-1).getSize());
	}

	@Override
	public boolean vertexDataIsUploaded() {
		return vertexDataIsUploaded;
	}

	@Override
	public void uniformScale(float scale) {
		modelMatrix.scale(new Vector3f(scale, scale, scale));
	}
	
}
