package modules.render;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.imageio.ImageIO;

import modules.loader.Scene;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.ARBFramebufferObject;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL14;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;
import org.lwjgl.opengl.Util;

import utility.FileIO;

/**
 * Encapsulation for one FBO.
 * 
 * @author DarkMatt
 *
 */

public class FBO {

	private int textureCount;
	private boolean hasDepth;
	private boolean bit32;
	private boolean mipMaps;
	private int totalTextCount;
	private int[] buffers;
	
	/** FBO index */
	private IntBuffer fbo_index = BufferUtils.createIntBuffer(1);
	/** Textures */
	private IntBuffer texture_index;
	/** Depth texture (?) */
	private IntBuffer depth_stencil_index = BufferUtils.createIntBuffer(1);
	
	private int sx, sy;
	
	private float sizeMultiplier;
	
	public FBO(int colorTextures, boolean depthTexture, int size_x, int size_y, float sizeMultiplier, boolean bit32, boolean mipMaps) {
		
		this.sizeMultiplier = sizeMultiplier;
		
		hasDepth = depthTexture;
		this.bit32 = bit32;
		this.mipMaps = mipMaps;
		textureCount = colorTextures;
		totalTextCount = textureCount + (depthTexture?1:0);
		texture_index = BufferUtils.createIntBuffer(totalTextCount);
		
		buffers = new int[totalTextCount];
		for (int i = 0; i < totalTextCount; i++) {
			buffers[i] = GL30.GL_COLOR_ATTACHMENT0+i;
		}
		
		init(size_x, size_y);
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("FBO: " + e);
		}
	}
	
	public void saveToFile(String path) {
		
		System.out.println("Writing to " + path + "albedo.png");
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));
		
		int bpp = 4; // Assuming a 32-bit display with a byte each for red, green, blue, and alpha.
		ByteBuffer buffer = BufferUtils.createByteBuffer(sx * sy * bpp);
		GL11.glReadPixels(0, 0, sx, sy, GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, buffer );
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
		
		
		File file = FileIO.getFile(path + "albedo.png"); // The file to save to.
		String format = "PNG"; // Example: "PNG" or "JPG"
		BufferedImage image = new BufferedImage(sx, sy, BufferedImage.TYPE_INT_RGB);
		 
		for(int x = 0; x < sx; x++) {
			for(int y = 0; y < sy; y++) {
				int i = (x + (sx * y)) * bpp;
				int r = buffer.get(i) & 0xFF;
				int g = buffer.get(i + 1) & 0xFF;
				int b = buffer.get(i + 2) & 0xFF;
				image.setRGB(x, sy - (y + 1), (0xFF << 24) | (r << 16) | (g << 8) | b);
			}
		}
		 
		try {
			ImageIO.write(image, format, file);
		} catch (IOException e) { e.printStackTrace(); }
		
		// ---------------------------------------------------------------------
		
		System.out.println("Writing to " + path + "normal.png");
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));
		
		GL11.glReadBuffer(GL30.GL_COLOR_ATTACHMENT0+1);
		
		bpp = 4; // Assuming a 32-bit display with a byte each for red, green, blue, and alpha.
		buffer = BufferUtils.createByteBuffer(sx * sy * bpp);
		GL11.glReadPixels(0, 0, sx, sy, GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, buffer );
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
		
		
		file = FileIO.getFile(path + "normal.png"); // The file to save to.
		format = "PNG"; // Example: "PNG" or "JPG"
		image = new BufferedImage(sx, sy, BufferedImage.TYPE_INT_RGB);
		 
		for(int x = 0; x < sx; x++) {
			for(int y = 0; y < sy; y++) {
				int i = (x + (sx * y)) * bpp;
				int r = buffer.get(i) & 0xFF;
				int g = buffer.get(i + 1) & 0xFF;
				int b = buffer.get(i + 2) & 0xFF;
				image.setRGB(x, sy - (y + 1), (0xFF << 24) | (r << 16) | (g << 8) | b);
			}
		}
		 
		try {
			ImageIO.write(image, format, file);
		} catch (IOException e) { e.printStackTrace(); }
		
		// ---------------------------------------------------------------------
		
		System.out.println("Writing to " + path + "light.png");
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));
		
		GL11.glReadBuffer(GL30.GL_COLOR_ATTACHMENT0+2);
		
		bpp = 4; // Assuming a 32-bit display with a byte each for red, green, blue, and alpha.
		buffer = BufferUtils.createByteBuffer(sx * sy * bpp);
		GL11.glReadPixels(0, 0, sx, sy, GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, buffer );
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
		
		
		file = FileIO.getFile(path + "light.png"); // The file to save to.
		format = "PNG"; // Example: "PNG" or "JPG"
		image = new BufferedImage(sx, sy, BufferedImage.TYPE_INT_RGB);
		 
		for(int x = 0; x < sx; x++) {
			for(int y = 0; y < sy; y++) {
				int i = (x + (sx * y)) * bpp;
				int r = buffer.get(i) & 0xFF;
				int g = buffer.get(i + 1) & 0xFF;
				int b = buffer.get(i + 2) & 0xFF;
				image.setRGB(x, sy - (y + 1), (0xFF << 24) | (r << 16) | (g << 8) | b);
			}
		}
		 
		try {
			ImageIO.write(image, format, file);
		} catch (IOException e) { e.printStackTrace(); }
		
		// ---------------------------------------------------------------------------
		
		System.out.println("Writing to " + path + "depth.png");
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));
		
		//GL11.glReadBuffer(GL11.GL_NONE);
		
		bpp = 4; // Assuming a 32-bit display with a byte each for red, green, blue, and alpha.
		buffer = BufferUtils.createByteBuffer(sx * sy * bpp);
		GL11.glReadPixels(0, 0, sx, sy, GL11.GL_DEPTH_COMPONENT, GL11.GL_FLOAT, buffer );
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
		
		file = FileIO.getFile(path + "depth.png"); // The file to save to.
		format = "PNG"; // Example: "PNG" or "JPG"
		image = new BufferedImage(sx, sy, BufferedImage.TYPE_INT_RGB);
		 
		for(int x = 0; x < sx; x++) {
			for(int y = 0; y < sy; y++) {
				int i = (x + (sx * y)) * bpp;
				int r = buffer.get(i) & 0xFF;
				int g = buffer.get(i + 1) & 0xFF;
				int b = buffer.get(i + 2) & 0xFF;
				image.setRGB(x, sy - (y + 1), (0xFF << 24) | (r << 16) | (g << 8) | b);
			}
		}
		 
		try {
			ImageIO.write(image, format, file);
		} catch (IOException e) { e.printStackTrace(); }
		
	}
	
	public float[] readData() {
		
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));
		
		FloatBuffer buf = BufferUtils.createFloatBuffer(sx*sy*4);
		//FloatBuffer buf = ByteBuffer.allocateDirect(sx * sy * 16).asFloatBuffer();
        GL11.glReadPixels(0, 0, sx, sy, GL11.GL_RGBA, GL11.GL_FLOAT, buf); //GL11.GL_RGBA
		
        FloatBuffer tmp = buf.duplicate();
        
        float[] out = new float[tmp.limit()];
        tmp.get(out);
        
        Display.update();
        
        GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
        
		return out;
		
	}
	
	public float[] readData(int mipMapLevel) {
		
		
		
		return null;
	}
	
	public int[] size() {
		return new int[]{sx, sy};
	}
	
	public void init(int size_x, int size_y) {
		
		//sx = (int)sizeMultiplier*size_x;
		//sy = /*(int)sizeMultiplier*/size_y;
		sx = (int)sizeMultiplier*size_x;// * Scene.GRIDDIMENSION[0];
		sy = (int)sizeMultiplier*size_y;// * Scene.GRIDDIMENSION[1];
		
		//CHECK version
		String version = GL11.glGetString(GL11.GL_VERSION);
		if (Integer.valueOf(version.substring(0,1)) < 3) {
			throw new RuntimeException("Pre OpenGL 3.0 do not support core FBO.");
		}
		
		//generate FBO
		GL30.glGenFramebuffers(fbo_index); 
		
		//texture image
		GL11.glGenTextures(texture_index);
		//renderbuffer image
		GL30.glGenRenderbuffers(depth_stencil_index);
		
		for (int i = 0; i < textureCount; i++) {

			//create texture 1-n
			GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture_index.get(i));
			if (bit32) {
				GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, 
						GL30.GL_RGB32F, sx, sy, 0, 
						GL11.GL_RGBA, GL11.GL_UNSIGNED_INT, (ByteBuffer) null);
			} else {
				GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, 
						GL11.GL_RGB, sx, sy, 0, 
						GL11.GL_RGBA, GL11.GL_UNSIGNED_INT, (ByteBuffer) null);
			}
			
			if (mipMaps) {
				//GL30.glGenerateMipmap(GL11.GL_TEXTURE_2D); // allocated space for mipmaps
			}
			
			
			if (mipMaps) {
				//GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL12.GL_TEXTURE_MAX_LEVEL, 2);
				//GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL14.GL_GENERATE_MIPMAP, GL11.GL_TRUE);
				GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR_MIPMAP_LINEAR); //GL_NEAREST_MIPMAP_NEAREST//GL_LINEAR_MIPMAP_LINEAR//GL_LINEAR
			} else {
				GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR); //GL_LINEAR
			}
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
			
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL14.GL_MIRRORED_REPEAT); //GL14.GL_MIRRORED_REPEAT GL13.GL_CLAMP_TO_BORDER
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL14.GL_MIRRORED_REPEAT);
		
			/*FloatBuffer color = BufferUtils.createFloatBuffer(4);
			color.put(new float[]{1.0f, 1.0f, 1.0f, 1.0f});
			color.position(0);
			GL11.glTexParameter(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_BORDER_COLOR, color);*/
			
		}
		if (hasDepth) {
			GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture_index.get(textureCount));
			GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, 
					GL30.GL_DEPTH24_STENCIL8, sx, sy, 0, //GL_DEPTH_COMPONENT24 old GL_DEPTH24_STENCIL8
					GL30.GL_DEPTH_STENCIL, GL30.GL_UNSIGNED_INT_24_8, (IntBuffer) null); //GL11.GL_DEPTH_COMPONENT old GL_DEPTH_STENCIL; GL_UNSIGNED_BYTE old GL30.GL_UNSIGNED_INT_24_8
			
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR); //GL_LINEAR or GL_NEAREST
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
			
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL13.GL_CLAMP_TO_BORDER);
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL13.GL_CLAMP_TO_BORDER);
			
			// border color for shadows
			FloatBuffer color = BufferUtils.createFloatBuffer(4);
			color.put(new float[]{1.0f, 1.0f, 1.0f, 1.0f});
			color.position(0);
			GL11.glTexParameter(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_BORDER_COLOR, color);
			
			//GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL14.GL_TEXTURE_COMPARE_MODE, GL30.GL_COMPARE_REF_TO_TEXTURE);
			//GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL14.GL_TEXTURE_COMPARE_FUNC, GL11.GL_LEQUAL);
		}
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("FBO: " + e);
		}
		
		GL30.glBindFramebuffer(
				GL30.GL_FRAMEBUFFER, 
				fbo_index.get(0));
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("FBO: " + e);
		}
		
		//Bind Depth
		GL30.glBindRenderbuffer(
				GL30.GL_RENDERBUFFER, 
				depth_stencil_index.get(0));
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("FBO: " + e);
		}
		
		// data storage
		GL30.glRenderbufferStorage(
				GL30.GL_RENDERBUFFER, 
				GL11.GL_DEPTH_COMPONENT, sx, sy); //TODO - new GL_DEPTH_COMPONENT old GL_DEPTH_STENCIL_ATTACHMENT (old throws invalid enum!)
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("FBO: " + e);
		}
		
		for (int i = 0; i < textureCount; i++) {
			//Bind FBO RENDERTARGET 1-n
			GL30.glFramebufferTexture2D(
					GL30.GL_FRAMEBUFFER, 
					GL30.GL_COLOR_ATTACHMENT0+i,
					GL11.GL_TEXTURE_2D, texture_index.get(i), 0);
		}
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("FBO: " + e);
		}
		
		if (hasDepth) {
			// image for depth buffer - moved down here - check
			GL30.glFramebufferRenderbuffer(
					GL30.GL_FRAMEBUFFER,
					GL30.GL_DEPTH_STENCIL_ATTACHMENT, //TODO - new GL_DEPTH_ATTACHMENT old GL_DEPTH_STENCIL_ATTACHMENT
					GL30.GL_RENDERBUFFER, depth_stencil_index.get(0));
			GL30.glFramebufferTexture2D(
					GL30.GL_FRAMEBUFFER, 
					GL30.GL_DEPTH_STENCIL_ATTACHMENT, //TODO - new GL_DEPTH_ATTACHMENT old GL_DEPTH_STENCIL_ATTACHMENT
					GL11.GL_TEXTURE_2D, texture_index.get(textureCount), 0);
		}
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("FBO: " + e);
		}
		
		//ERROR CHECKING
		checkError();
		
		//unbind
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
		GL30.glBindFramebuffer(
				ARBFramebufferObject.GL_FRAMEBUFFER, 
				0);
		
	}
	
	public void bindAll() {
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));

		IntBuffer ib = BufferUtils.createIntBuffer(totalTextCount);
		ib.put(buffers);
		ib.position(0);
		GL20.glDrawBuffers(ib);
		
		GL11.glViewport(0, 0, sx, sy);
		GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
	}
	
	/**
	 * Binds an atlas region depending on the defined grid
	 * 
	 * @param index
	 */
	public void bindAtlasRegion(int index) {
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));

		IntBuffer ib = BufferUtils.createIntBuffer(totalTextCount);
		ib.put(buffers);
		ib.position(0);
		GL20.glDrawBuffers(ib);
		
		// caluclate index
		int x = index % Scene.GRIDDIMENSION[0];
		int y = index / Scene.GRIDDIMENSION[0];
		
		GL11.glViewport(x * sx/Scene.GRIDDIMENSION[0], y * sy/Scene.GRIDDIMENSION[1], sx/Scene.GRIDDIMENSION[0], sy/Scene.GRIDDIMENSION[1]);
		
		//GL11.glViewport(index * sx/(Scene.NUMLIGHTS+1), 0, sx/(Scene.NUMLIGHTS+1), sy);
		//GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
	}
	
	/**
	 * Bind specific draw-buffers.
	 * 
	 * @param buffers
	 */
	public void bindCustomized(int[] bufferArray) {
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));
		
		IntBuffer ib = BufferUtils.createIntBuffer(bufferArray.length);
		ib.put(bufferArray);
		ib.position(0);
		GL20.glDrawBuffers(ib);
		
		GL11.glViewport(0, 0, sx, sy);
		//GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
	}
	
	/**
	 * Binds a single output for drawing.
	 * 
	 * @param index
	 */
	public void bindSingle(int index) {
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));
		
		GL11.glDrawBuffer(GL30.GL_COLOR_ATTACHMENT0 + index);
		
		GL11.glViewport(0, 0, sx, sy);
		//GL11.glClear(GL11.GL_COLOR_BUFFER_BIT);
	}
	
	public void clear() {
		//GL11.glViewport(0, 0, sx, sy);
		GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
	}
	
	/**
	 * Binds no output buffer - this is just for a stencil pass.
	 */
	public void bindNone() {
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, fbo_index.get(0));
		
		GL11.glDrawBuffer(GL11.GL_NONE);
		GL11.glReadBuffer(GL11.GL_NONE);
		
		GL11.glViewport(0, 0, sx, sy);
	}
	
	public void unbind() {
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
	}
	
	public void genMipMap() {
		for (int i = 0; i < textureCount; i++) {
			GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture_index.get(i));
			GL30.glGenerateMipmap(GL11.GL_TEXTURE_2D);
		}
	}
	
	public boolean hasMipMap() {
		return mipMaps;
	}
	
	public void bind_texture(int index) {
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture_index.get(index));
	}
	
	public void bind_depth_texture() {
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture_index.get(totalTextCount-1));
	}
	
	public void unbind_texture() {
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
	}
	
	public void blitFrom(int index) {
		GL30.glBindFramebuffer(GL30.GL_DRAW_FRAMEBUFFER, 0);
		GL30.glBindFramebuffer(GL30.GL_READ_FRAMEBUFFER, fbo_index.get(0));
		GL11.glReadBuffer(GL30.GL_COLOR_ATTACHMENT0+index);
		GL30.glBlitFramebuffer(0, 0, sx, sy, 
                0, 0, sx, sy, GL11.GL_COLOR_BUFFER_BIT, GL11.GL_LINEAR);
	}
	
	/*
	public void blitDepthToScreen() {
		GL30.glBindFramebuffer(GL30.GL_READ_FRAMEBUFFER, fbo_index.get(0));
		//GL11.glReadBuffer(GL30.GL_DEPTH_STENCIL_ATTACHMENT);
		GL30.glBindFramebuffer(GL30.GL_DRAW_FRAMEBUFFER, 0);
		//GL11.glDrawBuffer(GL11.GL_NONE);
		GL30.glBlitFramebuffer(0, 0, sx, sy, 0, 0, sx, sy, 
				GL11.GL_COLOR_BUFFER_BIT|GL11.GL_DEPTH_BUFFER_BIT, GL11.GL_NEAREST);
		
		checkError();
	}*/
	
	private void checkError() {
		int framebuffer = GL30.glCheckFramebufferStatus(GL30.GL_FRAMEBUFFER);
		switch ( framebuffer ) {
			case GL30.GL_FRAMEBUFFER_COMPLETE:
				break;
			case GL30.GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
				throw new RuntimeException( "FrameBuffer: " + fbo_index
						+ ", has caused a GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT exception" );
			case GL30.GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
				throw new RuntimeException( "FrameBuffer: " + fbo_index
						+ ", has caused a GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT exception" );
			case GL30.GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER:
				throw new RuntimeException( "FrameBuffer: " + fbo_index
						+ ", has caused a GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER exception" );
			case GL30.GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER:
				throw new RuntimeException( "FrameBuffer: " + fbo_index
						+ ", has caused a GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER exception" );
			default:
				throw new RuntimeException("Unexpected reply from glCheckFramebufferStatus: " + framebuffer);
		}
	}
	
}
