package modules.render;

import modules.loader.Scene;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL30;
import org.lwjgl.opengl.Util;

/**
 * Encapsulates a group of FBOs.
 * Currently has 2 output textures attached to the FBO and a depth.
 * (2 render-targets)
 * 
 * @author DarkMatt
 *
 */

public class FBOCollection {
	
	private FBO[][] fbos;
	
	public FBOCollection(int size_x, int size_y) {
		
		/**
		 * FBOs are filled:
		 * 
		 *  0 			 - screen FBO
		 *  1..1+numLS+1 - all albedo FBOs
		 *    			 - all normal FBOs
		 *    			 - all intensity FBOs
		 *    			 - all depth FBOs
		 * 
		 */
		
		/*int untilDepth = (1+Scene.NUMLIGHTS) * 3;
		int splitFbos = (1+Scene.NUMLIGHTS) * 4; // 4 fbos for every camera
		
		fbos = new FBO[splitFbos + 1 + Scene.NUMLIGHTS];
		
		fbos[0] = new FBO(3, true, size_x * Scene.GRIDDIMENSION[0], size_y * Scene.GRIDDIMENSION[1], 1.0f, false, false); // Shadow Depth FBO
		
		
		for (int i = 0; i < splitFbos; i++) {
			
			if (i<untilDepth) {
				//albedo, normal, intensity
				fbos[i+1] = new FBO(1, false, size_x, size_y, 1.0f, false, true);
			} else {
				//depth
				fbos[i+1] = new FBO(1, false, size_x, size_y, 1.0f, true, true);
			}
			
		}
		// result fbos
		for (int i = 0; i < Scene.NUMLIGHTS; i++) {
			fbos[i+splitFbos+1] = new FBO(1, false, size_x, size_y, 1.0f, false, false);
		}*/
		
		fbos = new FBO[2 + Scene.NUMLIGHTS][];
		
		// screen initial fbo
		fbos[0] = new FBO[] { new FBO(3, true, size_x * Scene.GRIDDIMENSION[0], size_y * Scene.GRIDDIMENSION[1], 1.0f, false, false) }; // Shadow Depth FBO
		
		for (int i = 0; i < 1 + Scene.NUMLIGHTS; i++) {
			
			fbos[1+i] = new FBO[6];
			
			//albedo
			fbos[1+i][0] = new FBO(1, false, size_x, size_y, 1.0f, false, true);
			//normal
			fbos[1+i][1] = new FBO(1, false, size_x, size_y, 1.0f, false, true);
			//intensity
			fbos[1+i][2] = new FBO(1, false, size_x, size_y, 1.0f, false, true);
			//depth
			fbos[1+i][3] = new FBO(1, false, size_x, size_y, 1.0f, true, true);
			
			// result
			fbos[1+i][4] = new FBO(1, false, size_x, size_y, 1.0f, false, false);
			
			// depth 2
			fbos[1+i][5] = new FBO(1, false, size_x, size_y, 2.0f, true, true);
		}
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("FBOCollection: " + e);
		}
		
	}
	
	public void reInit(int size_x, int size_y) {
		for (int i = 0; i < fbos.length; i++) {
			for (int j = 0; j < fbos.length; j++) {
				fbos[i][j].init(size_x, size_y);
			}
		}
	}
	
	public void saveToFile(int fbo, int type, String path) {
		fbos[fbo][type].saveToFile(path);
	}
	
	public float[] getPixelData(int fbo, int type) {
		return fbos[fbo][type].readData();
	}
	
	public int[] getSize(int fbo, int type) {
		return fbos[fbo][type].size();
	}
	
	/**
	 * @param index of the FBO
	 */
	public void bindAll(int fbo, int type) {
		fbos[fbo][type].bindAll();
	}
	
	public void bindAtlasRegion(int fbo, int type, int atlasIndex) {
		fbos[fbo][type].bindAtlasRegion(atlasIndex);
	}
	
	/**
	 * Bind a single draw-buffer to draw into.
	 * 
	 * @param index of the FBO
	 * @param bufferToBind the single drawBuffer to bind
	 */
	public void bindSingle(int fbo, int type, int bufferToBind) {
		fbos[fbo][type].bindSingle(bufferToBind);
	}
	
	public void clear(int fbo, int type) {
		fbos[fbo][type].clear();
	}
	
	public void genMipMaps(int fbo, int type) {
		fbos[fbo][type].genMipMap();
	}
	
	public void unbind() {
		GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
	}
	
	public void bind_texture(int fbo, int type, int textureUnit) {
		fbos[fbo][type].bind_texture(textureUnit);
	}
	
	public void unbind_texture() {
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
	}
	
}