package modules.render;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.FloatBuffer;
import java.util.HashMap;

import modules.render.objects.VBOArrayObject;

import org.lwjgl.opengl.GL20;

import utility.FileIO;

public class Shader {

	private int pId;
	
	private HashMap<String, Integer> shaderUniforms;
	
	public Shader(String path) {
		
		shaderUniforms = new HashMap<String, Integer>();
		
		int vsId = loadShader(path+".vert", GL20.GL_VERTEX_SHADER);
		int fsId = loadShader(path+".frag", GL20.GL_FRAGMENT_SHADER);
		
		pId = GL20.glCreateProgram();
		GL20.glAttachShader(pId, vsId);
		GL20.glAttachShader(pId, fsId);
		
		GL20.glBindAttribLocation(pId, VBOArrayObject.Location.DEFAULT_VERTEX_POSITION.ordinal(), "in_Position");
		GL20.glBindAttribLocation(pId, VBOArrayObject.Location.DEFAULT_COLOR_POSITION.ordinal(), "in_Color");
		GL20.glBindAttribLocation(pId, VBOArrayObject.Location.DEFAULT_TEXTCOORD_POSITION.ordinal(), "in_TextCoord");
		GL20.glBindAttribLocation(pId, VBOArrayObject.Location.DEFAULT_NORMAL_POSITION.ordinal(), "in_Normal");
		
		GL20.glLinkProgram(pId);
		
		String debugOut = GL20.glGetProgramInfoLog(pId, 5000);
		if (debugOut.length() > 0) {
			throw new RuntimeException(debugOut);
		}
		
	}
	
	private int loadShader(String filename, int type) {
		StringBuilder shaderSource = new StringBuilder();
		int shaderID = 0;
		 
		try {
			BufferedReader reader = new BufferedReader(FileIO.readFile(filename));
			String line;
			while ((line = reader.readLine()) != null) {
				shaderSource.append(line).append("\n");
			}
			reader.close();
		} catch (IOException e) {
			System.out.println("Could not read file.");
			e.printStackTrace();
		}
		
		shaderID = GL20.glCreateShader(type);
		GL20.glShaderSource(shaderID, shaderSource);
		GL20.glCompileShader(shaderID);
		 
		return shaderID;
	}
	
	public void bind() {
		GL20.glUseProgram(pId);
	}
	
	public void unbind() {
		GL20.glUseProgram(0);
	}
	
	public int initUniform(String name) {
		int location = 0;
		location = GL20.glGetUniformLocation(pId, name);
		
		shaderUniforms.put(name, location);
		
		return location;
	}
	
	public void setValue1i(String name, int value) {
		GL20.glUniform1i(shaderUniforms.get(name), value);
	}
	
	public void setValueVec2f(String name, float v1, float v2) {
		GL20.glUniform2f(shaderUniforms.get(name), v1, v2);
	}
	
	public void setValueVec3f(String name, float v1, float v2, float v3) {
		GL20.glUniform3f(shaderUniforms.get(name), v1, v2, v3);
	}
	
	public void setValueMat4f(String name, FloatBuffer value) {
		GL20.glUniformMatrix4(shaderUniforms.get(name), true, value);
	}
	
}
