package modules.render;

import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.imageio.ImageIO;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.EXTAbgr;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL12;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL30;

import utility.FileIO;


/**
 * Toolkit createImage
 *   or
 * ImageIO read
 * 
 * @author DarkMatt
 *
 */

public class TextureCollection {
	
	public TextureCollection() {
		
	}
	
	private static int nextPowerOfTwo(int number) {
		int n = 2;
		while (n <= number) {
			n*=2;
		}
		return n;
	}
	
	public int uploadImage(String path) {
        
		IntBuffer texBuff = BufferUtils.createIntBuffer(1);
		
		GL11.glGenTextures(texBuff);
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, texBuff.get(0));
		
		try {
			BufferedImage image = ImageIO.read(FileIO.readFIS(path));
			
			int width = image.getWidth();
			int height = image.getHeight();
			boolean hasAlpha = image.getColorModel().hasAlpha();
			
			int numCh = 3;
			if (hasAlpha) numCh = 4;
			
			WritableRaster raster = image.getRaster();
	        DataBufferByte buffer = (DataBufferByte) raster.getDataBuffer();
			byte[] data = buffer.getData();
			
			// WTF! ------
			int nextPowerOfTwo = nextPowerOfTwo(width);
			int counter = -1; 
			byte[] dataPad = new byte[numCh*nextPowerOfTwo*height];
			for (int j = 0; j < height; j++) {
				for (int i = 0; i < numCh*nextPowerOfTwo; i++) {
					if (i < numCh*width) {
						counter ++;
					}
					dataPad[i+numCh*nextPowerOfTwo*j] = data[counter];
				}
			}
			// WTF end
			
			ByteBuffer tex_data = BufferUtils.createByteBuffer(numCh*nextPowerOfTwo*height);
			tex_data.put(dataPad);
			tex_data.position(0);
			
			GL11.glPixelStorei(GL11.GL_UNPACK_ROW_LENGTH, nextPowerOfTwo);
			
			if (!hasAlpha) {
				GL11.glTexImage2D(
						GL11.GL_TEXTURE_2D, 0, 
						GL11.GL_RGB, width, height, 0, 
						GL12.GL_BGR, GL11.GL_UNSIGNED_BYTE, tex_data);
			} else {
				GL11.glTexImage2D(
						GL11.GL_TEXTURE_2D, 0, 
						GL11.GL_RGBA, width, height, 0, 
						EXTAbgr.GL_ABGR_EXT, GL11.GL_UNSIGNED_BYTE, tex_data);
			}
			
			GL30.glGenerateMipmap(GL11.GL_TEXTURE_2D);
			//GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_NEAREST);
			//GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR_MIPMAP_LINEAR);
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL11.GL_REPEAT);
			GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL11.GL_REPEAT);
	        
		} catch (IOException e) {
			e.printStackTrace();
		}
        
		//System.out.println("TextureCollection: image " + path + " loaded at " + texBuff.get(0) + ".");
		
		return texBuff.get(0);
		
	}
	
	public int uploadArray(int width, int height, float[] image) {
		
		GL13.glActiveTexture(GL13.GL_TEXTURE0);
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
		
		IntBuffer texBuff = BufferUtils.createIntBuffer(1);
		
		GL11.glGenTextures(texBuff);
		int index = texBuff.get(0);
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, index);
		
		int numCh = 4; // with alpha
		
		// WTF! ------
		int nextPowerOfTwo = nextPowerOfTwo(width);
		int counter = -1; 
		float[] dataPad = new float[numCh*nextPowerOfTwo*height];
		for (int j = 0; j < height; j++) {
			for (int i = 0; i < numCh*nextPowerOfTwo; i++) {
				if (i < numCh*width) {
					counter ++;
				}
				dataPad[i+numCh*nextPowerOfTwo*j] = image[counter];
			}
		}
		// WTF end
		
		FloatBuffer tex_data = BufferUtils.createFloatBuffer(numCh*nextPowerOfTwo*height);
		tex_data.put(dataPad);
		tex_data.position(0);
		
		GL11.glPixelStorei(GL11.GL_UNPACK_ROW_LENGTH, nextPowerOfTwo); //or height?
		
		/*GL11.glTexImage2D(
				GL11.GL_TEXTURE_2D, 0, 
				GL11.GL_RGBA, width, height, 0, 
				GL11.GL_RGBA, GL11.GL_FLOAT, tex_data);*/
		GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, 
				GL11.GL_RGBA, width, height, 
				0, GL11.GL_RGBA, GL11.GL_FLOAT, tex_data);
		
		//GL30.glGenerateMipmap(GL11.GL_TEXTURE_2D);
		GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);
		GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
		GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL11.GL_REPEAT);
		GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL11.GL_REPEAT);
		
		return index;
	}
	
	public void deleteTexture(int index) {
		GL11.glDeleteTextures(index);
	}
	
	public void bindTexture(int index) {
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, index);
	}
	
}
