package modules.render;


import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;

import modules.render.objects.VBOArrayObject;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.Util;

/**
 * 
 * @author DarkMatt
 *
 */

public class VBO {
	
	private ArrayList<IntBuffer> vbos;
	// stores data per object and per data entry
	private ArrayList<ArrayList<VBOArrayObject>> vertexData;
	
	public VBO() {
		vbos = new ArrayList<IntBuffer>();
		vertexData = new ArrayList<ArrayList<VBOArrayObject>>();
	}
	
	
	public int init(ArrayList<VBOArrayObject> vertexData) {
		
		int size = vertexData.size();
		
		IntBuffer vboBuffer = BufferUtils.createIntBuffer(size);
		GL15.glGenBuffers(vboBuffer);
		
		// vertex data
		for (int i = 0; i < size; i++) {
			
			VBOArrayObject obj = vertexData.get(i);
			
			// float data
			if (obj.getData() instanceof FloatBuffer) {
				GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboBuffer.get(i));
				GL15.glBufferData(GL15.GL_ARRAY_BUFFER, (FloatBuffer)obj.getData(), GL15.GL_STATIC_DRAW);
			}
			
			// init data
			if (obj.getData() instanceof IntBuffer) {
				GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, vboBuffer.get(i));
				GL15.glBufferData(GL15.GL_ELEMENT_ARRAY_BUFFER, (IntBuffer)obj.getData(), GL15.GL_STATIC_DRAW);
			}
			
			
		}
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, 0);
		
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("VBO: " + e);
		}
		
		vbos.add(vboBuffer);
		
		// add VBO data (size, location, etc)
		ArrayList<VBOArrayObject> toAdd = new ArrayList<VBOArrayObject>();
		for (int i = 0; i < vertexData.size(); i++) {
			toAdd.add(vertexData.get(i).copyWithoutBuffer());
		}
		this.vertexData.add(toAdd);
		
		return vbos.size()-1;
	}
	
	/**
	 * Old basic render call per object
	 * 
	 * @param objectId
	 * @param type
	 * @param indicesSize
	 */
	public void draw(int objectId, int type, int indicesSize) {
		
		for (int i = 0; i < vertexData.get(objectId).size()-1; i++) {
			GL20.glEnableVertexAttribArray(vertexData.get(objectId).get(i).getLocation()); // can be defined in the shader
			GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vbos.get(objectId).get(i));	//vbo pos
			GL20.glVertexAttribPointer(vertexData.get(objectId).get(i).getLocation(), vertexData.get(objectId).get(i).getSize(), vertexData.get(objectId).get(i).getType(), false, 0, 0);
		}
		
		int vbos_at_last_element = vbos.get(objectId).capacity()-1;
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, vbos.get(objectId).get(vbos_at_last_element)); 	//for indices
			
		GL11.glDrawElements(type, indicesSize, GL11.GL_UNSIGNED_INT, 0);
			
		try { 
			Util.checkGLError();
		} catch (Exception e) {
			System.out.println("VBO: " + e);
		}
		
		for (int i = 0; i < vertexData.get(objectId).size(); i++) {
			GL20.glDisableVertexAttribArray(vertexData.get(objectId).get(i).getLocation());
		}
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, 0);
		
	}
	
	/**
	 * Delete
	 */
	public void deleteAll() {
		for (int i = 0; i < vbos.size(); i++) {
			GL15.glDeleteBuffers(vbos.get(i));
		}
	}
	
	/*@Deprecated
	public void update(int objectId, short[] indices, float[] vertices) {
		
		ShortBuffer ind = BufferUtils.createShortBuffer(indices.length);
		ind.put(indices);
		ind.position(0);
		
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, vboIBO.get(objectId));
		GL15.glBufferData(GL15.GL_ELEMENT_ARRAY_BUFFER, ind, GL15.GL_STATIC_DRAW);
		
		FloatBuffer in = BufferUtils.createFloatBuffer(vertices.length);
		in.put(vertices);
		in.position(0);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(5*objectId+0));
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, in, GL15.GL_STREAM_DRAW);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		
	}
	
	@Deprecated
	public void updatePositions(int objectId, float[] vertices) {
		
		
		FloatBuffer in = BufferUtils.createFloatBuffer(vertices.length);
		in.put(vertices);
		in.position(0);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(5*objectId+0));
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, in, GL15.GL_STREAM_DRAW);
		//GL15.glBufferSubData(GL15.GL_ARRAY_BUFFER, 0, in); //(long)(4*start)
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		
	}
	
	@Deprecated
	public void updateColors(int objectId, float[] color) {
		
		FloatBuffer in = BufferUtils.createFloatBuffer(color.length);
		in.put(color);
		in.position(0);
		
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboId.get(5*objectId+1));
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, in, GL15.GL_STREAM_DRAW);
		//GL15.glBufferSubData(GL15.GL_ARRAY_BUFFER, 0, in); //(long)(4*start)
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);
		
	}*/
		
}
