package modules.render;

import java.awt.Canvas;
import java.awt.Frame;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.JFrame;

import modules.commands.CommandExecutor;
import modules.commands.ICommand;

import org.lwjgl.LWJGLException;
import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;

/**
 * Wrapper for the window and resize.
 * 
 * TODO - problem resize:
 *  - viewport => result strangely distorted
 *  - no viewport - image still small
 * 
 * @author DarkMatt
 *
 */

public class WindowWrapper {

	public static final float WINDOWSCALE = 2.0f;
	public static final int WINDOWX = 256; //64 //840
	public static final int WINDOWY = 256;
	
	private static final boolean FULLSCREEN = false;
	private static final String WINDOWNAME = "Engine";
	
	private static final int MAXX = Display.getDesktopDisplayMode().getWidth();
	private static final int MAXY = Display.getDesktopDisplayMode().getHeight();
	
	private Canvas canvas;
	private int px, py;
	private float sx, sy;
	private float aspect;
	
	private boolean resized = false;
	
	private JFrame frame;
	private Canvas c;
	private ComponentListener cl = new ComponentListener() {
		public void componentResized(ComponentEvent e) {
			System.out.println("resized");
		}
		public void componentShown(ComponentEvent arg0) {}
		public void componentMoved(ComponentEvent arg0) {}
		public void componentHidden(ComponentEvent arg0) {}
	};
	private WindowListener wl = new WindowListener() {
		public void windowClosing(WindowEvent arg0) {
			commandExecutor.execute(shutDownCommand);
		}
		public void windowActivated(WindowEvent arg0) {}
		public void windowClosed(WindowEvent arg0) {}
		public void windowDeactivated(WindowEvent e) {}
		public void windowDeiconified(WindowEvent arg0) {}
		public void windowIconified(WindowEvent arg0) {}
		public void windowOpened(WindowEvent arg0) {}
	};
	
	private CommandExecutor commandExecutor;
	private ICommand shutDownCommand;
	
	
	public WindowWrapper(CommandExecutor commandExecutor, ICommand shutDownCommand) {
		this.commandExecutor = commandExecutor;
		this.shutDownCommand = shutDownCommand;
		this.canvas = initWindow();

		resize();
	}
	
	/**
	 * Swing specific
	 * 
	 * @return
	 */
	private Canvas initWindow() {
		
		frame = new JFrame();
		if (!FULLSCREEN) {
			// plus window borders... //THERE IS A MINIUMUM SIZE
			frame.setBounds(100, 100, (int)(WINDOWX*WINDOWSCALE) + 16, (int)(WINDOWY*WINDOWSCALE) + 38); //frame.setBounds(100, 100, 840, 580);
		} else {
			frame.setExtendedState(Frame.MAXIMIZED_BOTH);
			frame.setUndecorated(true);
		}
		//frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		frame.setVisible(true);
		frame.setFocusable(true);
		c = new Canvas();
		c.setFocusable(true);
		c.setIgnoreRepaint(true);
		frame.getContentPane().add(c);
		
		frame.addComponentListener(cl);
		frame.addWindowListener(wl);
		
		return c;
	}
	
	/**
	 * Don't use OpenGL here.
	 */
	public void resize() {
		this.px = canvas.getParent().getWidth();
		this.py = canvas.getParent().getHeight();
		
		if (FULLSCREEN) {
			this.px = MAXX;
			this.py = MAXY;
		}
		
		/*
		 * ad aspect:
		 * parameterization of the screen in -1 to 1 along the minimal direction
		 */
		aspect = (float)px/(float)py;
		if (aspect > 1.0) {
			this.sx = aspect;
			this.sy = 1.0f;
		} else {
			this.sx = 1.0f;
			this.sy = 1.0f/aspect;
		}
		
		resized = true;
	}
	
	/**
	 * Called in initialize. (render thread)
	 */
	public void initializeDisplay() {
		
		// set canvas size to window size
		canvas.setSize(canvas.getParent().getSize());
		canvas.requestFocus();
		
		try {
			Display.setParent(canvas);
		} catch (LWJGLException e) {
			e.printStackTrace();
		}
		
		//takes a while... (attention)
		Display.setTitle(WINDOWNAME);
		try {
			Display.setFullscreen(FULLSCREEN);
			Display.setDisplayMode(new DisplayMode(px, py));
			Display.create();
		} catch (LWJGLException e) {
			e.printStackTrace();
		}
		
		Mouse.setGrabbed(true);
		
	}
	
	public void updateResize() {
		if (resized) {
			try {
				Display.setFullscreen(FULLSCREEN);
			} catch (LWJGLException e) {
				e.printStackTrace();
			}
			GL11.glViewport(0, 0, px, py);
			resized = false;
		}
	}
	
	/**
	 * Comments:
	 *  - focus needed
	 *  - Display.update() needed
	 */
	public void updateKeys() {
		
		while(Keyboard.next()) {
			
			if (Keyboard.getEventKey() == Keyboard.KEY_ESCAPE) {
				commandExecutor.execute(shutDownCommand);
			} else {
				int key = Keyboard.getEventKey();
				if (Keyboard.isKeyDown(key)) {
					String value = Keyboard.getKeyName(key);
					System.out.print(value);
				}
			}
			
		}
		
		//System.out.println(Mouse.getDX() +"/"+ Mouse.getDY());
	}
	
	public float[] getScreenXY() {
		return new float[]{sx, sy};
	}
	
	public int[] getPixelXY() {
		return new int[]{px, py};
	}
	
	public void shutdown() {
		frame.dispose();
	}
	
}
