package modules.render.cpu;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

/**
 * Conversion from depth to position image plane.
 * 
 * @author DarkMatt
 *
 * @param <A>
 * @param <N>
 * @param <D>
 * @param <L>
 */

public class ImagePlanePosition<A, N, D, L> {

	private int sx, sy;
	
	private ImagePlane<A, N, D, L> base;
	private Vector3f[][] ws_position;
	
	public ImagePlanePosition(ImagePlane<A, N, D, L> base) {
		this.base = base;
		this.sx = base.getSx();
		this.sy = base.getSy();
	}
	
	private Vector3f getWorldSpacePosition(Vector2f textCoord) {

		Vector2f relCoord = new Vector2f((float)textCoord.x/(float)(sx-1), (float)textCoord.y/(float)(sy-1));
		
		int[] coord = new int[]{(int)textCoord.x, (int)textCoord.y};//new int[]{(int)(textCoord.x*(c_size[0]-1)), ((int)textCoord.y*(c_size[1]-1))};
		
		float depth = (Float)base.Depth(coord[0], coord[1]);//texture2D(depthAtlas[0], textCoord).x;
		Vector4f screenSpacePos = new Vector4f(relCoord.x, relCoord.y, depth, 1.0f);
	    screenSpacePos.x = 2.0f*(screenSpacePos.x-0.5f);
	    screenSpacePos.y = 2.0f*(screenSpacePos.y-0.5f);
	    screenSpacePos.z = 2.0f*(screenSpacePos.z-0.5f);
	    screenSpacePos.w = 2.0f*(screenSpacePos.w-0.5f);
	    Vector4f worldSpace = new Vector4f(screenSpacePos);
	    Matrix4f.transform(base.getInvProjMatrix(), screenSpacePos, worldSpace); //invProjMatrix
	    Matrix4f.transform(base.getInvViewMatrix(), worldSpace, worldSpace); //invViewMatrix
	    worldSpace.x /= worldSpace.w;
	    worldSpace.y /= worldSpace.w;
	    worldSpace.z /= worldSpace.w;
	    worldSpace.w = 1.0f;
	    //Vector4f worldSpace = invViewMatrix * invProjMatrix * screenSpacePos;
	    //worldSpace / worldSpace.w;
	    return new Vector3f(worldSpace.x, worldSpace.y, worldSpace.z);
	}
	
	public void calculateWorldPositions() {
		ws_position = new Vector3f[sx][sy];
		
		for (int ly = 0; ly < sy; ly++) {
			for (int lx = 0; lx < sx; lx++) {
				
				ws_position[lx][ly] = getWorldSpacePosition(new Vector2f(lx, ly));
				
			}
		}
		
	}
	
	public Vector3f getWs_position(int x, int y) {
		return ws_position[x][y];
	}
	
	public Vector3f[][] getWs_positionArray() {
		return ws_position;
	}

}
