package modules.render.cpu.boundingsphere;


import org.lwjgl.util.vector.Vector3f;

public class AABBBoundingSphere extends AbstractBoundingSphereGenerator {
	
	@Override
	public void calculateBoundingSphere() {
		
		if (boudingSpheresInput.size() > 0) {
		
			// generate bounding-box first
			
			float[] min = new float[]{Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE};
			float[] max = new float[]{-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE};
			
			BoundingSphere[] minMaxSphere = new BoundingSphere[6];
			
			for (int i = 0; i < boudingSpheresInput.size(); i++) {
				BoundingSphere b = boudingSpheresInput.get(i);
				
				if (b.getPos().x - b.getRadius() < min[0]) { // x minimum
					min[0] = b.getPos().x - b.getRadius();
					minMaxSphere[0] = b;
				}
				if (b.getPos().y - b.getRadius() < min[1]) { // y minimum
					min[1] = b.getPos().y - b.getRadius();
					minMaxSphere[1] = b;
				}
				if (b.getPos().z - b.getRadius() < min[2]) { // z minimum
					min[2] = b.getPos().z - b.getRadius();
					minMaxSphere[2] = b;
				}
				
				if (b.getPos().x + b.getRadius() > max[0]) { // x maximum
					max[0] = b.getPos().x + b.getRadius();
					minMaxSphere[3] = b;
				}
				if (b.getPos().y + b.getRadius() > max[1]) { // y maximum
					max[1] = b.getPos().y + b.getRadius();
					minMaxSphere[4] = b;
				}
				if (b.getPos().z + b.getRadius() > max[2]) { // z maximum
					max[2] = b.getPos().z + b.getRadius();
					minMaxSphere[5] = b;
				}
			}
			
			// generate a sphere from the bounding-box
			
			boundingSphereResult = new BoundingSphere();
			
			Vector3f pos = new Vector3f((min[0]+max[0])/2.0f, (min[1]+max[1])/2.0f, (min[2]+max[2])/2.0f);
			boundingSphereResult.setPos(pos);
			
			float radius = 0.0f;
			for (int i = 0; i < minMaxSphere.length; i++) {
				float newRadius = Vector3f.sub(pos, minMaxSphere[i].getPos(), null).length() + minMaxSphere[i].getRadius();
				if (newRadius > radius) radius = newRadius;	
			}
			boundingSphereResult.setRadius(radius);
			
			// error check (necessary)
			for (int i = 0; i < boudingSpheresInput.size(); i++) {
				float dist = Vector3f.sub(boundingSphereResult.getPos(), boudingSpheresInput.get(i).getPos(), null).length();
				if (dist > boundingSphereResult.getRadius() + boudingSpheresInput.get(i).getRadius()) {
					//System.out.println("Error bounding-sphere outside at " + boudingSpheresInput.get(i).getPos());
					boundingSphereResult.setRadius(dist-boudingSpheresInput.get(i).getRadius());
				}
			}
			
			// final redundant error check
			for (int i = 0; i < boudingSpheresInput.size(); i++) {
				float dist = Vector3f.sub(boundingSphereResult.getPos(), boudingSpheresInput.get(i).getPos(), null).length();
				if (dist > boundingSphereResult.getRadius() + boudingSpheresInput.get(i).getRadius()) {
					System.out.println("Error bounding-sphere outside at " + boudingSpheresInput.get(i).getPos());
				}
			}
			
		} else {
			
			boundingSphereResult = new BoundingSphere();
			
		}
		
	}
	
}
