package modules.render.cpu.boundingsphere;

import org.lwjgl.util.vector.Vector3f;

public class BoundingSphere {

	private Vector3f pos;
	private float radius; // 0 for one pixel
	
	/** generation with parameters */
	public BoundingSphere(Vector3f pos, float radius) {
		this.pos = pos;
		this.radius = radius;
	}
	
	public BoundingSphere() {
		pos = new Vector3f(0.0f, 0.0f, 0.0f);
		radius = 0.0f;
	}
	
	public BoundingSphere(BoundingSphere bs) {
		pos = bs.getPos();
		radius = bs.getRadius();
	}
	
	/** Doesn't work if spheres intersect */
	public BoundingSphere(BoundingSphere b1, BoundingSphere b2) {
		
		// mean between 2 center-points and radius offset
		
		Vector3f p1 = b1.getPos();
		Vector3f p2 = b2.getPos();
		
		Vector3f dir = Vector3f.sub(p2, p1, null).normalise(null); // from p1 to p2
		
		Vector3f offset2 = (Vector3f) new Vector3f(dir).scale(b2.getRadius());
		Vector3f offset1 = (Vector3f) new Vector3f(dir).negate().scale(b1.getRadius());
		
		Vector3f.add(p1, offset1, p1);
		Vector3f.add(p2, offset2, p2);
		
		// mean
		pos = (Vector3f) Vector3f.add(p1, p2, null).scale(0.5f);
		radius = Vector3f.sub(pos, p1, null).length();
	}
	
	public Vector3f getPos() {
		return pos;
	}
	
	public void setPos(Vector3f pos) {
		this.pos = pos;
	}
	
	public float getRadius() {
		return radius;
	}
	
	public void setRadius(float radius) {
		this.radius = radius;
	}

}
