package utility;

/**
 * Timer implementation for frame-rate independence and FPS display.
 * 
 * @author DarkMatt
 *
 */

public class Timer {
	
	private float dt;
	private long currentTime;
	private long current;
	
	//average over dt (idea backup: game engine design)
	private final int DTSAMPLESIZE = 20;
	private double[] dtSamples = new double[DTSAMPLESIZE];
	private int dtSampleCounter = 0;
	
	private final int MAX_SAPLES = 50;
	private int sample_counter = 0;
	private long fpsMeasure;
	private String fps = "";
	
	private long waitFor;
	private long waitForMeasure;
	private boolean timerActive;
	
	private boolean everStarted;
	
	public Timer() {
		timerActive = false;
		everStarted = false;
	}
	
	public boolean timerActive() {
		return timerActive;
	}
	
	public void waitFor(long ms) {
		waitFor = ms * 1000000;
		waitForMeasure = System.nanoTime() + waitFor;
		timerActive = true;
		everStarted = true;
	}
	
	public boolean everStarted() {
		return everStarted;
	}
	
	public boolean waitedEnough() {
		if (System.nanoTime() > waitForMeasure) {
			waitForMeasure = System.nanoTime() + waitFor;
			timerActive = false;
			return true;
		} else
			return false;
	}
	
	//every loop
	public void startMeasure() {
		currentTime = System.nanoTime(); //currentTimeMillis
		if (sample_counter < MAX_SAPLES-1) sample_counter ++;
		else {
			fpsMeasure = currentTime;
			sample_counter = 0;
		}
	}
	
	//every loop
	public void endMeasure() {
		//System.out.println((System.currentTimeMillis() - currentTime));
		//average over dt
		current = System.nanoTime() - currentTime;
		if (++dtSampleCounter < DTSAMPLESIZE) {
		} else {
			dtSampleCounter = 0;
		}
		dtSamples[dtSampleCounter] = current/1000000000.0;
		for (int i=0; i<DTSAMPLESIZE; i++) {
			dt += dtSamples[i];
		}
		dt /= DTSAMPLESIZE;
		//dt = current/1000000000.0f;
		//framerate display
		if (sample_counter == MAX_SAPLES-1) {
			double c = (System.nanoTime() - fpsMeasure)/MAX_SAPLES;
			
			String val = String.valueOf( 1000000000./c );
			int comma = val.indexOf('.');
			fps = String.valueOf( 1000000000./c ).substring(0, comma);
		}
	}
	
	public double dt() {
		
		//if (10000.0 * dt > 1.0)
		//	System.out.println(10000.0 * dt);
		
		return dt;
	}
	
	public String getFPS() {
		return fps;
	}
	
}
